#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Launch and monitor local shell jobs."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2024, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import os
import subprocess
import sys

import sciexp2.expdef.system
import sciexp2.common.instance
from sciexp2.common import progress
from sciexp2.common.utils import execute_with_sigint


class System (sciexp2.expdef.system.System):
    """Manage local shell jobs."""

    STATE_CMD = "ps xwww"

    ASSUMES = []
    DEFINES = []

    def compute_state(self):
        # build instance group of job states
        self._jobs = sciexp2.common.instance.InstanceGroup()
        for instance in self._launchers:
            job = Job(self,
                      sciexp2.expdef.system.Job.compute_state(self, instance),
                      instance)
            self._jobs.add(job)


class Job (sciexp2.expdef.system.Job):
    """A local shell script job."""

    def state(self):
        state = self["_STATE"]
        if state == sciexp2.expdef.system.Job.DONE:
            name = self._system.get_relative_path(self["DONE"])
        elif state == sciexp2.expdef.system.Job.FAILED:
            name = self._system.get_relative_path(self["FAIL"])
        elif state in [sciexp2.expdef.system.Job.NOTRUN, sciexp2.expdef.system.Job.OUTDATED]:
            name = self._system.get_relative_path(self["LAUNCHER"])
        else:
            raise ValueError("Unknown job state: %r" % state)
        return state, name

    def submit(self, *args):
        launcher = os.sep.join([self._system._base, self["LAUNCHER"]])
        assert os.path.isfile(launcher)

        cmd = ["bash"] + self._submit_args(args) + [launcher]
        progress.verbose(" %s", " ".join(cmd))
        kwargs = dict(stdout=sciexp2.expdef.system._DEVNULL,
                      stderr=sciexp2.expdef.system._DEVNULL)
        if progress.level() > progress.LVL_PROGRESS:
            kwargs["stderr"] = progress.get_writer(sys.stderr)
        if progress.level() >= progress.LVL_VERBOSE:
            kwargs["stdout"] = progress.get_writer(sys.stdout)

        res = execute_with_sigint(cmd, **kwargs)
        if res != 0:
            raise subprocess.CalledProcessError(res, cmd)

    def kill(self, *args):
        raise Exception("Cannot kill local shell script jobs")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Parse and apply filter expressions."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2024, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections.abc
import re
import linecache


def _re_match(value, pattern):
    cre = re.compile(pattern)
    return cre.match(value) is not None


class Filter:
    """Boolean expression to check against a dict-like object.

    The filter contains an arbitrary Python expression, where every variable
    will be taken from the dict we are matching the filter against.


    Parameters
    ----------
    expression : Filter or dict or str, optional
        Expression to use in the filter.


    Raises
    ------
    SyntaxError
        The expression is not valid.


    Notes
    -----
    If `expression` is a dict-like object, it will define an expression that
    exactly matches its items.

    Every filter will have the following global names defined:

    ============================  ==========================
    ``re_match(var, str)``        Check if ``var`` matches the regular
                                  expression ``str``.
    ============================  ==========================


    See also
    --------
    validate
    match
    and_filters, or_filters


    Examples
    --------
    Filters can be easily composed together:

    >>> f1 = Filter("a < 3")
    >>> f2 = Filter("b == 4")
    >>> and_filters(f1, f2)
    Filter("(a < 3) and (b == 4)")
    >>> or_filters(f1, f2)
    Filter("(a < 3) or (b == 4)")

    Filter objects can be later matched against dict-like objects:

    >>> f = Filter("a < 3 and b == 4")
    >>> f.match(dict(a=2, b=4))
    True
    >>> f.match(dict(a=3, b=4))
    False

    Using a dict as an expression is equivalent to building a perfect match for
    the dict's items:

    >>> Filter({"VAR1": 1, "VAR2": 2})
    Filter("VAR1 == 1 and VAR2 == 2")

    """

    _GLOBALS = {"re_match": _re_match}

    def __init__(self, expression=None):
        if expression is None or expression == "":
            expression = "True"
        elif isinstance(expression, Filter):
            # pylint: disable=protected-access
            expression = expression._expression
        elif isinstance(expression, collections.abc.Mapping):
            keys = sorted(expression.keys())
            expression = " and ".join([f"{key} == {repr(expression[key])}"
                                       for key in keys])
        self._expression = expression
        self._code_id = f"<dynamic-{id(self._expression)}>"
        self._code = compile(self._expression, self._code_id, "eval")
        linecache.cache[self._code_id] = (len(self._expression), None,
                                          self._expression.split("\n"),
                                          self._code_id)

    def __del__(self):
        if self._code_id in linecache.cache:
            del linecache.cache[self._code_id]

    def __str__(self):
        """Return a string representation of the filter."""
        return self._expression

    def __repr__(self):
        return f"Filter(\"{str(self)}\")"

    def validate(self, allowed):
        """Validate that variables in the filter are present in the given set.

        Parameters
        ----------
        allowed : set of variable names
            Set of variable names to allow on the filter.

        Raises
        ------
        NameError
            Filter contains a variable name not present in `allowed`.

        """
        present = set(self._code.co_names)
        missing = present - (set(allowed) | set(["re_match"]))
        if missing:
            missing = list(missing)
            raise NameError(f"name {repr(missing[0])} is not allowed")

    def match(self, source):
        """Check if the given `source` matches this filter.

        Parameters
        ----------
        source : dict-like
            Dictionary to match this filter against.

        Returns
        -------
        bool : Whether the match is positive or not.

        Raises
        ------
        NameError
            Filter contains a variable name not present in `source`.

        See also
        --------
        validate

        """
        # pylint: disable=eval-used
        return eval(self._code, dict(source), self._GLOBALS)


def and_filters(*filters):
    """Convenience function to *and* all `filters` together."""
    filters = [f"({Filter(f)})" for f in filters]
    expression = " and ".join(filters)
    return Filter(expression)


def or_filters(*filters):
    """Convenience function to *or* all `filters` together."""
    filters = [f"({Filter(f)})" for f in filters]
    expression = " or ".join(filters)
    return Filter(expression)


__all__ = [
    "Filter", "and_filters", "or_filters",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Automate generation of parametrized launchers."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2025, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections
import contextlib
import concurrent
import functools
import itertools
import operator
import os
import shutil
import sys
import threading
import warnings

from sciexp2.common import progress
from sciexp2.common import utils
from sciexp2.common.instance import InstanceGroup, Instance
from sciexp2.common.filter import *
from sciexp2.common.utils import execute_with_sigint
from sciexp2.common import pp
from sciexp2.common import text
from . import templates
from . import launcher


class _sorted_dict(object):
    """Helper class to print dicts with sorted keys."""
    def __init__(self, object):
        self._object = object
    def __repr__(self):
        keys = sorted(self._object.keys())
        return "{%s}" % ", ".join("%r: %r" % (key, self._object[key])
                                  for key in keys)

def get_executor_workers(parallel):
    if not parallel:
        return 1
    elif parallel is True:
        return os.cpu_count()
    else:
        return parallel


######################################################################
# Experiments

def _is_dummy(experiments):
    if isinstance(experiments, Experiments):
        return (len(experiments._experiments) == 1 and
                len(experiments._experiments.get_index(0)) == 0)
    else:
        return len(experiments) == 1 and len(experiments[0]) == 0


def _val_is_translate(val):
    return (isinstance(val, str) and
            len(text.get_variables(val)) > 0) or \
            isinstance(val, _WithExp) or \
            isinstance(val, _WithExpTpl)


def _get_translate_exp(exp):
    res = utils.OrderedSet()
    for key, vals in exp.items():
        if isinstance(vals, str):
            if _val_is_translate(vals):
                res.add(key)
        elif isinstance(vals, collections.abc.Iterable):
            if any(_val_is_translate(val) for val in vals):
                res.add(key)
    return res

def _get_translate(experiments):
    translate = utils.OrderedSet()
    for exp in experiments:
        translate |= _get_translate_exp(exp)
    return translate


def _do_execute(experiments, cmd, **kwargs):
    if len(experiments) == 0:
        experiments_list = [Instance()]
    else:
        experiments_list = experiments

    if isinstance(cmd, str):
        is_shell = True
        cmd = [cmd]
    else:
        is_shell = False

    parallel = kwargs.pop("parallel", None)
    repeat = kwargs.pop("repeat", "warn")
    if repeat not in ["ignore", "warn", "error"]:
        raise ValueError("invalid repeat value: %r" % repeat)
    std_args = {
        'stdin':  kwargs.pop("stdin",  None),
        'stdout': kwargs.pop("stdout", None),
        'stderr': kwargs.pop("stderr", None),
    }
    utils.assert_kwargs(kwargs)

    def std_translator(name):
        arg = std_args[name]
        if isinstance(arg, str):
            return text.Translator(arg)
        else:
            return None
    std_translator = {
        'stdin':  std_translator('stdin'),
        'stdout': std_translator('stdout'),
        'stderr': std_translator('stderr'),
    }


    # flatten command to get faster translation
    cmd_arg_sep = "|;argsep;|"
    cmd_flat = cmd_arg_sep.join(cmd)
    def get_cmd(cmd):
        return cmd.split(cmd_arg_sep)

    translator = text.Translator(cmd_flat)
    exp_cmds = translator.translate_many(experiments_list, with_envs=True)

    def execute(exp_cmd, progr):
        cmd = get_cmd(exp_cmd[0])
        exp = exp_cmd[1][0]
        exp_rest = exp_cmd[1][1:]

        progress.verbose(" ".join(cmd))
        if len(exp_rest) > 0:
            if repeat == "warn":
                warnings.warn(
                    "Repeated command translation for experiments %r and %r" % (
                        exp, exp_rest[0]))
            elif repeat == "error":
                raise RuntimeError(
                    "repeated command translation for experiments %r and %r" % (
                        exp, exp_rest[0]))


        def get_stdfile(name, env):
            @contextlib.contextmanager
            def as_context(arg):
                yield arg

            translator = std_translator[name]
            if translator is None and name == "stdout":
                return progress.get_writer(sys.stdout)
            elif translator is None and name == "stderr":
                return progress.get_writer(sys.stderr)
            elif translator is None:
                return as_context(std_args[name])
            else:
                with env.auto_translate():
                    path = translator.translate(env)
                return open(path, "w")

        with get_stdfile("stdin",  exp) as stdin,\
             get_stdfile("stdout", exp) as stdout,\
             get_stdfile("stderr", exp) as stderr:
            res = execute_with_sigint(
                *cmd,
                shell=is_shell,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr)
        if res != 0:
            raise Exception("Command did not finish correctly: "
                            "%s -> %d" % (" ".join(cmd), res))

        return len(exp_cmd[1])

    # use progr() explicitly to ensure progress is made after each command
    with progress.get(experiments_list, msg="Executing commands...") as progr:
        max_workers = get_executor_workers(parallel)

        def execute_with_progress(exp):
            count = execute(exp, progr)
            progr(count)

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            list(executor.map(execute_with_progress, exp_cmds))


def _do_recombine(experiments, exp_filters, variables, var_filters, append):
    exp_filters = and_filters(*exp_filters)
    var_filters = and_filters(*var_filters)

    # normalize variable value types
    for key, val in variables.items():
        if not isinstance(val, collections.abc.Iterable) or \
           isinstance(val, str):
            variables[key] = [val]
        elif isinstance(val, collections.abc.Sized):
            variables[key] = list(val)

    # calculate various properties about variable values
    var_translate = _get_translate_exp(variables)
    var_from_func = [var for var, vals in variables.items()
                     if any(isinstance(val, _FromFunctionKey) for val in vals)]
    var_size = functools.reduce(operator.mul,
                                (len(vals) for vals in variables.values()),
                                1)

    # validate variable filters
    var_filters.validate((set(experiments.variables()) | set(variables.keys())) - var_translate)

    new_experiments = InstanceGroup()
    new_dups = set()

    total_size = len(experiments) * var_size
    if append:
        total_size += var_size

    def cartesian_product(base, variables_missing=None):
        """Yield cartesian product of `variables` and overlay each on `base`"""
        if variables_missing is None:
            variables_missing = sorted(variables.keys())

        if len(variables_missing) == 0:
            yield {}

        else:
            key = variables_missing[0]
            variables_missing = variables_missing[1:]

            for val in variables[key]:
                for exp in cartesian_product(base, variables_missing):
                    new_exp = Instance(base)
                    new_exp.update(exp)
                    new_exp[key] = val
                    yield new_exp

    def cartesian_product_ff(base, ff_info, ffs_missing=None):
        """Yield cartesian product of `ff_info` and overlay each on `base`"""
        if ffs_missing is None:
            ffs_missing = sorted(ff_info.keys())

        if len(ffs_missing) == 0:
            yield {}

        else:
            ff = ffs_missing[0]
            ffs_missing = ffs_missing[1:]

            for val in ff_info[ff][0]:
                for exp in cartesian_product_ff(base, ff_info, ffs_missing):
                    new_exp = Instance(base)
                    new_exp.update(exp)
                    for var in ff_info[ff][1]:
                        new_exp[var[0]] = val[var[1]]
                    yield new_exp

    def merge_old_experiment(exp, force=False):
        exp_dups = tuple(exp.items())
        if exp_dups not in new_dups:
            new_experiments.add(exp)
            new_dups.add(exp_dups)
        elif force:
            new_experiments.add(exp)

    def merge_new_experiment_noff(progr, exp):
        if var_filters.match(exp):
            merge_old_experiment(exp)
        progr()

    def merge_new_experiment(progr, exp):
        ff_vars = dict(((var, exp[var]) for var in var_from_func
                        if isinstance(exp[var], _FromFunctionKey)))

        if len(ff_vars) == 0:
            merge_new_experiment_noff(progr, exp)

        else:
            exp = dict((key, val) for key, val in exp.items()
                       if key not in ff_vars)

            ff_info = {}
            progr_size_add = 1
            for var, val in ff_vars.items():
                val = val._from_func
                if val not in ff_info:
                    exp_ff = dict(exp)
                    vals = val._get_results(exp)
                    if len(vals) == 0:
                        # experiment dropped, but still increase progress
                        progr()
                        return
                    ff_info[val] = (vals, [])
                    progr_size_add *= len(vals)
                ff_info[val][1].append((var, ff_vars[var]._key_name))

            assert progr_size_add > 0
            progr += progr_size_add

            for exp_prod in cartesian_product_ff(exp, ff_info):
                merge_new_experiment_noff(progr, exp_prod)


    with progress.get(total_size,
                      msg="recombining variables...") as progr:

        if append:
            for exp in experiments:
                merge_old_experiment(exp, force=True)
                progr(var_size)

            for exp_prod in cartesian_product({}):
                merge_new_experiment(progr, exp_prod)
        else:
            for exp_base in experiments:
                try:
                    is_match = exp_filters.match(exp_base)
                except NameError:
                    is_match = False

                if is_match:
                    for exp_prod in cartesian_product(exp_base):
                        merge_new_experiment(progr, exp_prod)
                else:
                    merge_old_experiment(exp_base)
                    progr(var_size)

    return new_experiments, var_translate


def _do_pack(experiments, template_from, template_to, dereference):
    if len(experiments) == 0:
        experiments_list = [Instance()]
    else:
        experiments_list = experiments
    if dereference is None:
        dereference = experiments.dereference

    translate_from = text.Translator(template_from)
    translate_to = text.Translator(template_to)

    done_first = {}
    done_exp = {}
    with progress.get(len(experiments_list), msg="packing files...") as progr:
        for exp in experiments_list:
            with exp.auto_translate():
                path_from = utils.get_path(translate_from.translate(exp))
                path_to = utils.get_path(translate_to.translate(exp))
                path_to = os.path.join(experiments.out, path_to)

                done_idx = path_from + "+" + path_to
                if done_idx in done_first:
                    if done_first[done_idx]:
                        done_first[done_idx] = False
                        progress.info(
                            "Skipping already packed file: %s -> %s",
                            path_from, path_to)
                    progress.verbose(
                        "Conflicting experiments during pack:\n%s\n%s",
                        done_exp[done_idx], exp)
                else:
                    done_first[done_idx] = True
                    done_exp[done_idx] = exp

                    progress.debug("Packing: %s -> %s", path_from, path_to)
                    utils.copy_path(path_from, path_to, dereference=dereference)

                progr()


def _do_translate(experiments, exp_filters, exp_translate,
                  templates, with_exps, with_unique):
    exp_filters = and_filters(*exp_filters)
    parsed_templates = [text.Translator(template)
                        for template in templates]

    if with_exps:
        if with_unique:
            translations = collections.OrderedDict()
            if len(templates) == 1:
                def add(translated, exp):
                    translated = translated[0]
                    if translated not in translations:
                        translations[translated] = []
                    translations[translated].append(dict(exp))
            else:
                def add(translated, exp):
                    if translated not in translations:
                        translations[translated] = []
                    translations[translated].append(dict(exp))
            def result():
                return translations.items()
        else:
            translations = []
            if len(templates) == 1:
                def add(translated, exp):
                    translations.append((translated[0], [dict(exp)]))
            else:
                def add(translated, exp):
                    translations.append((translated, [dict(exp)]))
            def result():
                return translations
    else:
        if with_unique:
            translations = utils.OrderedSet()
            if len(templates) == 1:
                def add(translated, exp):
                    translations.add(translated[0])
            else:
                def add(translated, exp):
                    translations.add(tuple(translated))
            def result():
                return translations
        else:
            translations = []
            if len(templates) == 1:
                def add(translated, exp):
                    translations.append(translated[0])
            else:
                def add(translated, exp):
                    translations.append(tuple(translated))
            def result():
                return translations

    for exp in experiments.select(exp_filters):
        with exp.auto_translate():
            try:
                is_match = exp_filters.match(exp)
            except NameError:
                is_match = False

            if is_match:
                _WithExp.set_experiment(exp)
                _WithExpTpl.set_experiment(exp)
                translated = [template.translate(exp)
                              for template in parsed_templates]
                add(translated, exp)

    return list(result())


def _do_generate(experiments, template_from, template_to,
                 pre_exp_fun, post_exp_fun, parallel):
    out = os.path.abspath(utils.get_path(experiments.out))
    translate_from = text.Translator(template_from)
    translate_to = text.Translator(template_to)

    done_set_lock = threading.Lock()
    done_set = {}
    class done_state:
        def __init__(self):
            self.lock = threading.Lock()
            self.is_first = True
            self.is_second = False
            self.first_exp = None
        def set_first(self, exp):
            with self.lock:
                is_first = self.is_first
                if is_first:
                    self.is_first = False
                    self.first_exp = exp
                    is_second = True
                else:
                    is_second = False
            return is_first, is_second
    def get_done(path_from, path_to):
        key = (path_from, path_to)
        with done_set_lock:
            if key not in done_set:
                done_set[key] = done_state()
            return done_set[key]

    with progress.get(len(experiments), msg="generating files...") as progr:

        def generate_with_progress(exp_info):
            idx, exp = exp_info
            with exp.auto_translate():
                _WithExp.set_experiment(exp)
                _WithExpTpl.set_experiment(exp)

                path_from = utils.get_path(translate_from.translate(exp))
                path_to = utils.get_path(translate_to.translate(exp))
                path_to = os.path.join(out, path_to)

                done = get_done(path_from, path_to)
                is_first, is_second = done.set_first(exp)
                if not is_first:
                    msg = f"WARNING: Skipping already generated file: {path_from} -> {path_to}"
                    msg += "\n"
                    msg += f"Source file template: {template_from}"
                    msg += "\n"
                    msg += f"Destination file template: {template_to}"
                    msg += "\n"
                    msg += f"Conflicting experiments:\n{done.first_exp}\n{exp}"
                    if is_second:
                        progress.info(msg)
                    else:
                        progress.verbose(msg)
                else:
                    progress.debug(f"Generating: {path_from} -> {path_to}")

                    exp, contents_to = pre_exp_fun(idx, exp, path_from, path_to)

                    utils.assert_dir(os.path.dirname(path_to))
                    if os.path.isfile(path_to):
                        with open(path_to, "r") as f:
                            contents_to_old = f.read()
                    else:
                        contents_to_old = ""
                        utils.get_file(path_to, "w")
                    shutil.copymode(path_from, path_to)

                    if contents_to != contents_to_old:
                        with open(path_to, "w") as f:
                            f.write(contents_to)

                    post_exp_fun(idx, exp, path_to)

            progr()

        max_workers = get_executor_workers(parallel)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            list(executor.map(generate_with_progress, enumerate(experiments)))


def _do_generate_user(experiments, template_from, template_to, parallel,
                      backend, backend_options):
    @functools.lru_cache(16)
    def get_translator(path_from):
        file_from = text.TemplatePath(path_from)
        return text.Translator(file_from, backend=backend,
                               backend_options=backend_options)
    def pre_exp_fun(idx, exp, path_from, path_to):
        translator = get_translator(path_from)
        return exp, translator.translate(exp)
    def post_exp_fun(idx, exp, path_to):
        pass
    _do_generate(experiments, template_from, template_to,
                 pre_exp_fun, post_exp_fun, parallel)

def _do_generate_jobs(experiments, system, template_to, export, depends,
                      submit_args, job_desc, parallel):
    template = templates.get(system)

    depends = set(depends) | set(["{{LAUNCHER}}"])

    # export all variables in depends
    export = set(export)
    for elem in depends:
        if not isinstance(elem, str):
            raise TypeError("depends element is not a string: %r" % elem)
        for elem_var in text.get_variables(elem, nested="all"):
            if elem_var == "LAUNCHER":
                continue
            export.add(elem_var)

    if isinstance(experiments, ExperimentsView):
        experiments_vars = list(experiments._base._experiments.variables())
    else:
        experiments_vars = list(experiments._experiments.variables())
    for export_elem in export:
        if not isinstance(export_elem, str):
            raise TypeError("export element is not a string: %r" % export_elem)
        if export_elem not in experiments_vars:
            raise ValueError("missing variable to export: %s" % export_elem)
    export = export | set(template.system.assumes())

    out = os.path.abspath(utils.get_path(experiments.out))
    job_desc = os.path.join(out, job_desc)
    job_desc = os.path.abspath(utils.get_path(job_desc))
    if not job_desc.startswith(out):
        raise ValueError("job descriptor file must be inside output directory")
    job_desc_base = os.path.dirname(job_desc[len(out)+1:])

    # load template file and apply overrides
    path_from = template.template_path
    with open(path_from, "r") as f:
        template_from = f.read()
    overrides = dict(template.overrides)
    for var in text.get_variables(template_from):
        if var not in overrides:
            overrides[var] = "{{%s}}" % var
    template_from = text.translate(text.TemplatePath(template.template_path),
                                   overrides, recursive=False)
    template_from_vars = text.get_variables(template_from)
    translate_from = text.Translator(text.TemplateContents(template_from))

    def pre_exp_fun(idx, exp, path_from, path_to):
        exp = dict(exp)

        if "LAUNCHER" in exp:
            raise ValueError("experiment has reserved variable 'LAUNCHER'")
        exp["LAUNCHER"] = path_to[len(out)+1:]

        for key, val in template.defaults.items():
            if key not in exp:
                exp[key] = val

        contents_to = translate_from.translate(exp)

        return exp, contents_to

    experiments_generated = [None] * len(experiments)

    def post_exp_fun(idx, exp, path_to):
        for var in export:
            if var not in exp:
                continue
            val = exp[var]
            if not _val_is_translate(val):
                continue
            exp[var] = text.translate(val, exp)
        experiments_generated[idx] = Instance(exp)

    _do_generate(experiments, template.template_path, template_to,
                 pre_exp_fun, post_exp_fun, parallel)

    experiments_generated = InstanceGroup(
        (exp for exp in experiments_generated if exp is not None))

    launcher.save(job_desc, job_desc_base,
                  template.system, experiments_generated,
                  export, depends, submit_args)


class Experiments(pp.Pretty):
    """Define and generate experiment sets.

    Each experiment corresponds to a `dict` that maps variable names to their
    values, and can be accessed as if `Experiments` was a list.

    Parameters
    ----------
    experiments : sequence of dict, optional
        Initial experiment set.
    out : optional
        Initial value for the `out` attribute.
    dereference : bool, optional
        Initial value for the `dereference` attribute.

    Attributes
    ----------
    out : str
        Output directory for all non-absolute paths.
    dereference : bool
        Whether to dereference source symlinks during `pack`.

    See also
    --------
    with_exp, with_exp_tpl
        Use functions as variable values.
    from_function, from_find_files
        Use functions to create new variable values.
    sciexp2.common.text.translate
        Used to interpret template arguments in the `Experiments` methods, used
        with the variable maps of each experiment.
    sciexp2.common.text.extract
        Used to interpret template arguments in the `Experiments` methods, used
        to build the variable maps of new experiments.
    sciexp2.common.filter.Filter
        Values accepted for the filter arguments of the methods in
        `Experiments`. A list of strings will be used as an argument to
        `sciexp2.common.filter.and_filters`.

    """

    #: Output directory for all non-absolute paths.
    out = "./out"

    #: Whether to dereference source symlinks during pack.
    dereference = False

    def __init__(self, experiments=None, **kwargs):
        out = kwargs.pop("out", None)
        dereference = kwargs.pop("dereference", None)
        utils.assert_kwargs(kwargs)

        if experiments is None:
            experiments = InstanceGroup()
            # add dummy experiment
            experiments += Instance()
        if isinstance(experiments, Experiments) or\
           isinstance(experiments, ExperimentsView):
            if dereference is None:
                dereference = experiments.dereference
            if out is None:
                out = experiments.out
            self._experiments = experiments._experiments
            self._translate = utils.OrderedSet(experiments._translate)
        else:
            self._experiments = InstanceGroup()
            dups = set()
            for exp in experiments:
                exp_dups = tuple(exp.items())
                if exp_dups not in dups:
                    self._experiments.add(Instance(exp))
                    dups.add(exp_dups)
            self._translate = _get_translate(self._experiments)

        if out is not None:
            self.out = out
        if dereference is not None:
            self.dereference = dereference


    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            have_prev = False
            if len(self) > 0:
                p.pretty(list(map(lambda elem: _sorted_dict(elem), self)))
                have_prev = True
            if self.out != Experiments.out:
                if have_prev:
                    p.text(",")
                    p.breakable()
                p.text("out=%r" % self.out)
                have_prev = True
            if self.dereference != Experiments.dereference:
                if have_prev:
                    p.text(",")
                    p.breakable()
                p.text("dereference=%r" % self.dereference)


    def __iter__(self):
        if _is_dummy(self):
            return iter([])
        else:
            return iter(self._experiments)

    def __len__(self):
        if _is_dummy(self):
            return 0
        else:
            return len(self._experiments)

    def __getitem__(self, index):
        if _is_dummy(self):
            return list()[index]
        else:
            return self._experiments.get_index(index)

    def __eq__(self, other):
        return isinstance(other, Experiments) and \
            self.out == other.out and \
            self.dereference == other.dereference and \
            list(self._experiments) == list(other._experiments) and \
            list(self._translate) == list(other._translate)

    def view(self, *filters):
        """Get a view to a portion of this of this experiment set.

        Any changes to the experiments on a view will be reflected on the base
        experiment set and all its views. Adding new experiments to a view will
        maintain the relative order between experiments on the base experiment
        set.

        Parameters
        ----------
        filters : list of filters
            Filters to select a matching subset of the experiments.

        Examples
        --------
        A view shows a subset of the experiments:

        >>> e = Experiments()
        >>> e.params(a=range(2), b=range(2))
        >>> e
        Experiments([{'a': 0, 'b': 0},
                     {'a': 0, 'b': 1},
                     {'a': 1, 'b': 0},
                     {'a': 1, 'b': 1}])
        >>> v = e.view("a == b")
        >>> v
        ExperimentsView([{'a': 0, 'b': 0}, {'a': 1, 'b': 1}])

        All operations are performed only on selected elements, but apply to the
        underlying experiment set as if they happened "in-place":

        >>> v = e.view("a == b")
        >>> v.params(c=range(2))
        >>> v
        ExperimentsView([{'a': 0, 'b': 0, 'c': 0},
                         {'a': 0, 'b': 0, 'c': 1},
                         {'a': 1, 'b': 1, 'c': 0},
                         {'a': 1, 'b': 1, 'c': 1}])
        >>> e
        Experiments([{'a': 0, 'b': 0, 'c': 0},
                     {'a': 0, 'b': 0, 'c': 1},
                     {'a': 0, 'b': 1},
                     {'a': 1, 'b': 0},
                     {'a': 1, 'b': 1, 'c': 0},
                     {'a': 1, 'b': 1, 'c': 1}])

        For convenience, you can also get the inverse of a view, which is a view
        with the negated filters:

        >>> i = v.view_inverse()
        >>> i
        ExperimentsView([{'a': 0, 'b': 1}, {'a': 1, 'b': 0}])
        >>> i.params(d=range(2))
        >>> e
        Experiments([{'a': 0, 'b': 0, 'c': 0},
                     {'a': 0, 'b': 0, 'c': 1},
                     {'a': 0, 'b': 1, 'd': 0},
                     {'a': 0, 'b': 1, 'd': 1},
                     {'a': 1, 'b': 0, 'd': 0},
                     {'a': 1, 'b': 0, 'd': 1},
                     {'a': 1, 'b': 1, 'c': 0},
                     {'a': 1, 'b': 1, 'c': 1}])

        Finally, you can also create new views from other views:

        >>> n = v.view("c == 0")
        >>> n.params(e=[10, 20])
        >>> n.view_inverse().params(e=[30, 40])
        >>> e
        Experiments([{'a': 0, 'b': 0, 'c': 0, 'e': 10},
                     {'a': 0, 'b': 0, 'c': 0, 'e': 20},
                     {'a': 0, 'b': 0, 'c': 1, 'e': 30},
                     {'a': 0, 'b': 0, 'c': 1, 'e': 40},
                     {'a': 0, 'b': 1, 'd': 0},
                     {'a': 0, 'b': 1, 'd': 1},
                     {'a': 1, 'b': 0, 'd': 0},
                     {'a': 1, 'b': 0, 'd': 1},
                     {'a': 1, 'b': 1, 'c': 0, 'e': 10},
                     {'a': 1, 'b': 1, 'c': 0, 'e': 20},
                     {'a': 1, 'b': 1, 'c': 1, 'e': 30},
                     {'a': 1, 'b': 1, 'c': 1, 'e': 40}])

        """
        return ExperimentsView(self, Filter("True"), Filter(*filters))

    def execute(self, cmd, **kwargs):
        """Execute one program for each experiment on the set.

        If `cmd` is a single string, it's interpreted as a shell command,
        otherwise it's interpreted as a program name and its argument list.

        Arguments `stdin`, `stdout`, `stderr`, and all elements in `cmd` are
        interpreted as templates to translate with each experiment.

        Parameters
        ----------
        cmd : list of str
            Command to execute.
        repeat : str, optional
            What to do when more than one experiment is translated into the same
            command. Can be ``"ignore"``, ``"warn"`` (default), or ``"error"``.
        stdin, stdout, stderr : file or str, optional
            Paths for the standard input/output/error for the command (defaults
            to None).
        parallel : optional
            Whether to execute commands in parallel (default is `False`).

        Notes
        -----
        Command execution is internally handled by `subproces.Popen` and thus
        the `stdin`, `stdout` and `stderr` arguments can also have any value
        accepted by it.

        See also
        --------
        sciexp2.common.parallel.get_parallelism
            Used to interpret argument `parallel`.

        """
        _do_execute(self, cmd, **kwargs)


    def params(self, *filters, **variables):
        """Add new parameter permutations to experiment set.

        Performs the cartesian product between the current experiment set and
        all permutations of the given variable values (i.e., a parameter can
        have a sequence of values).

        Parameters
        ----------
        filters : list of filters, optional
            Ensures new experiments match the given filters (default is to
            accept all).
        variables : dict
            A dictionary of variable name / values pairs to perform
            permutations.

        Notes
        -----
        Values in `variables` can be templates referencing other variables
        (either existing in the experiment set or provided in `variables`), and
        will thus be translated accordingly.

        Providing a variable in `params` that already exists in the experiment
        set will override its value.

        Examples
        --------
        Can easily create value permutations that match the given filters:
        >>> e = Experiments()
        >>> e.params("a % 2 == 0",
        ...          a=range(4),
        ...          b=["foo", "bar"])
        >>> e
        Experiments([{'a': 0, 'b': 'foo'},
                     {'a': 0, 'b': 'bar'},
                     {'a': 2, 'b': 'foo'},
                     {'a': 2, 'b': 'bar'}])

        New variables can be appended to the experiment set:

        >>> e.params(c=3)
        >>> e
        Experiments([{'a': 0, 'b': 'foo', 'c': 3},
                     {'a': 0, 'b': 'bar', 'c': 3},
                     {'a': 2, 'b': 'foo', 'c': 3},
                     {'a': 2, 'b': 'bar', 'c': 3}])

        Including ones with multiple values:

        >>> e.params(d=[1, 2])
        >>> e
        Experiments([{'a': 0, 'b': 'foo', 'c': 3, 'd': 1},
                     {'a': 0, 'b': 'foo', 'c': 3, 'd': 2},
                     {'a': 0, 'b': 'bar', 'c': 3, 'd': 1},
                     {'a': 0, 'b': 'bar', 'c': 3, 'd': 2},
                     {'a': 2, 'b': 'foo', 'c': 3, 'd': 1},
                     {'a': 2, 'b': 'foo', 'c': 3, 'd': 2},
                     {'a': 2, 'b': 'bar', 'c': 3, 'd': 1},
                     {'a': 2, 'b': 'bar', 'c': 3, 'd': 2}])

        New values override old ones, even if that means eliminating some experiments:

        >>> e.params(b="baz")
        >>> e
        Experiments([{'a': 0, 'b': 'baz', 'c': 3, 'd': 1},
                     {'a': 0, 'b': 'baz', 'c': 3, 'd': 2},
                     {'a': 2, 'b': 'baz', 'c': 3, 'd': 1},
                     {'a': 2, 'b': 'baz', 'c': 3, 'd': 2}])

        """
        new_experiments = _do_recombine(self._experiments, [],
                                        variables, filters,
                                        False)
        self._experiments = new_experiments[0]
        self._translate |= new_experiments[1]


    def params_append(self, *filters, **variables):
        """Same as `params`, but appends the new permutations.

        Examples
        --------
        >>> e = Experiments()
        >>> e.params(a=range(2))
        >>> e.params_append(b=range(2))
        >>> e
        Experiments([{'a': 0},
                     {'a': 1},
                     {'b': 0},
                     {'b': 1}])

        """
        new_experiments = _do_recombine(self._experiments, [],
                                        variables, filters,
                                        True)
        self._experiments = new_experiments[0]
        self._translate |= new_experiments[1]


    def pack(self, template_from, template_to, dereference=None):
        """Copy files, accepting templates for both source and destination paths.

        Parameters
        ----------
        template_from, template_to : str
            Source/destination file templates.
        dereference : bool, optional
            Whether to dereference symlinks in source files (defaults to
            ``self.dereference``).

        Notes
        -----
        The copy is performed only if the source is newer than the destination,
        or if the destination does not exist.

        There can be cases where multiple source files correspond to a single
        destination file (after translating the templates). Such cases are
        logged by a message.

        """
        _do_pack(self, template_from, template_to, dereference)

    def generate(self, template_from, template_to,
                 parallel=True, backend="pystache", backend_options={}):
        """Generate files from other templated files.

        Like `pack`, the paths to source and destination files are
        templates. Every source file is, in addition, considered a template that
        will be translated with each of the experiments.

        Parameters
        ----------
        template_from, template_to : str
            Templates for Source/destination file paths.
        parallel : optional
            Whether to generate files in parallel (default is `True`).
        backend : {'pystache', 'jinja2'}, optional
            Templating backend to process the contents of `template_from`.

        Notes
        -----
        Like `pack`, if multiple source files correspond to a single destination
        file (after translating the templates), the case is logged by a message
        and only the first generated file is kept.

        """
        _do_generate_user(self, template_from, template_to,
                          parallel=parallel, backend=backend,
                          backend_options=backend_options)

    def generate_jobs(self, system, template_to, export=[], depends=[],
                      submit_args=[], job_desc="jobs.jd",
                      parallel=True):
        """Generate job files for given job system.

        Works similar to `generate`, but takes a predefined source file template
        that is identified by `system`. It also creates a job descriptor file
        that can be later used by the `launcher` program.

        Parameters
        ----------
        system : str
            Job system name.
        template_to : str
            Template to path of destination job scripts.
        export : list of str, optional
            List of variable names to export into the job descriptor file.
        depends : list of str, optional
            File path templates for job dependencies.
        submit_args : list of str, optional
            Templates for additional arguments to the job submission command.
        job_desc : str, optional
            Name of the output job descriptor file.
        parallel : optional
            Whether to generate files in parallel (default is `True`).

        All job systems implicitly define the *LAUNCHER* variable, which has the
        path of the to the generated script to run that. All job systems also
        define *DONE* and *FAIL* as file paths to identify the state of each
        job.

        The variables selected by `export` will be later available when using
        `sciexp2.expdef.launcher` with this job descriptor. Any variable
        directly or indirectly used in `depends` will be implicitly added to
        `export`. Variable *LAUNCHER* is also implicitly added.

        The templates specified by `depends` will be translated for each job
        info file paths, and each file path will be considered a dependency of
        that job. The path specified by *LAUNCHER* is implicitly added as a
        dependency.

        The state for a given job is decided as follows:

        - Not run: Neither the *DONE* nor the *FAIL* files exist.
        - Failed: The *FAIL* file exists.
        - Outdated: The *DONE* file exists, but is older than any of its
          dependencies.
        - Done: The *DONE* file exists, and is newer than any of its
          dependencies.
        - Running: Some job systems can track when a job is currently executing.


        Notes
        -----

        None of the experiments can define the *LAUNCHER* variable. Some job
        systems also define their own set of reserved variables and default
        values to others (in case your experiments don't specify them); see the
        job system's documentation for more information.

        """
        _do_generate_jobs(self, system, template_to, export, depends,
                          submit_args, job_desc, parallel)


    def translate(self, template, with_exps=False, with_unique=True):
        """Translate the given templates with each experiment.

        Parameters
        ----------
        template : str
            Template to translate.
        with_exps : bool, optional
            Return the list of experiments used to get each translation.
        with_unique : bool, optional
            Only return unique translations.

        Returns
        -------
        A list of the template translations (``with_exps=False``), or a list of
        tuples with each translation and the corresponding experiments
        (``with_exps=True``).

        Examples
        --------
        Translations are unique by default:
        >>> e = Experiments()
        >>> e.params(a=range(2), b=range(2), c=range(2))
        >>> e.translate("{{a}}-{{b}}")
        ['0-0', '0-1', '1-0', '1-1']
        >>> e.translate("{{a}}-{{b}}", with_unique=False)
        ['0-0', '0-0', '0-1', '0-1', '1-0', '1-0', '1-1', '1-1']

        And can be easily extended with the experiments that led to each
        translation:
        >>> e.translate("{{a}}-{{b}}", with_exps=True) == [
        ...     ('0-0', [{'a': 0, 'b': 0, 'c': 0}, {'a': 0, 'b': 0, 'c': 1}]),
        ...     ('0-1', [{'a': 0, 'b': 1, 'c': 0}, {'a': 0, 'b': 1, 'c': 1}]),
        ...     ('1-0', [{'a': 1, 'b': 0, 'c': 0}, {'a': 1, 'b': 0, 'c': 1}]),
        ...     ('1-1', [{'a': 1, 'b': 1, 'c': 0}, {'a': 1, 'b': 1, 'c': 1}])]
        True
        >>> e.translate("{{a}}-{{b}}", with_exps=True, with_unique=False) == [
        ...     ('0-0', [{'a': 0, 'b': 0, 'c': 0}]),
        ...     ('0-0', [{'a': 0, 'b': 0, 'c': 1}]),
        ...     ('0-1', [{'a': 0, 'b': 1, 'c': 0}]),
        ...     ('0-1', [{'a': 0, 'b': 1, 'c': 1}]),
        ...     ('1-0', [{'a': 1, 'b': 0, 'c': 0}]),
        ...     ('1-0', [{'a': 1, 'b': 0, 'c': 1}]),
        ...     ('1-1', [{'a': 1, 'b': 1, 'c': 0}]),
        ...     ('1-1', [{'a': 1, 'b': 1, 'c': 1}])]
        True

        """
        return _do_translate(self._experiments, [], self._translate,
                             [template], with_exps, with_unique)


######################################################################
# Views

class ExperimentsView (pp.Pretty):
    """Proxy to a subset of elements in a `Experiments` instance.

    Notes
    -----
    You can use a view with Python's ``with`` statement to improve code
    readability:

    >>> e = Experiments()
    >>> e.params(a=range(2), b=range(2))
    >>> with e.view("a != b") as v, v.view_inverse() as i:
    ...     v.params(c=1)
    ...     i.params(c=2)
    >>> e
    Experiments([{'a': 0, 'b': 0, 'c': 2},
                 {'a': 0, 'b': 1, 'c': 1},
                 {'a': 1, 'b': 0, 'c': 1},
                 {'a': 1, 'b': 1, 'c': 2}])

    """

    def __init__(self, experiments, filter_base, filter_top,
                 **kwargs):
        assert len(kwargs) == 0
        self._base = experiments
        self._filters = [filter_base, filter_top]
        for filter_ in self._filters:
            filter_.validate(set(self._base._experiments.variables()))

    def _get_experiments(self):
        filter_ = self._as_filter()
        return self._base._experiments.select(filter_, allow_unknown=True)

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            have_prev = False
            if len(self) > 0:
                p.pretty(list(map(lambda elem: _sorted_dict(elem), self._get_experiments())))
                have_prev = True
            if self.out != Experiments.out:
                if have_prev:
                    p.text(",")
                    p.breakable()
                p.text("out=%r" % self.out)
                have_prev = True
            if self.dereference != Experiments.dereference:
                if have_prev:
                    p.text(",")
                    p.breakable()
                p.text("dereference=%r" % self.dereference)

    def __iter__(self):
        return self._get_experiments()

    def __len__(self):
        return len(list(self._get_experiments()))

    def __getitem__(self, index):
        return list(self._get_experiments())[index]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _as_filter(self):
        return and_filters(*self._filters)

    @property
    def out(self):
        return self._base.out

    @property
    def dereference(self):
        return self._base.dereference

    def view(self, *filters):
        """Create a new view from this one.

        See also
        --------
        Experiments.view

        """
        return ExperimentsView(self._base,
                               self._as_filter(),
                               and_filters(*filters))

    def view_inverse(self):
        """Get the inverse view of this one.

        If using nested views, this will invert the filter of this one, not all
        filters for parent views.

        See also
        --------
        Experiments.view

        """
        return self._base.view(and_filters(self._filters[0],
                                           "not (%s)" % self._filters[1]))

    def execute(self, *args, **kwargs):
        """Same as `Experiments.execute`"""
        return _do_execute(self, *args, **kwargs)

    def params(self, *filters, **variables):
        """Same as `Experiments.params`"""
        new_experiments = _do_recombine(self._base._experiments, self._filters,
                                        variables, filters,
                                        False)
        self._base._experiments = new_experiments[0]
        self._base._translate |= new_experiments[1]

    def params_append(self, *filters, **variables):
        """Same as `Experiments.params_append`"""
        new_experiments = _do_recombine(self._base._experiments, [],
                                        variables, filters,
                                        True)
        self._base._experiments = new_experiments[0]
        self._base._translate |= new_experiments[1]

    def pack(self, template_from, template_to, dereference=None):
        """Same as `Experiments.pack`"""
        _do_pack(self, template_from, template_to, dereference=None)

    def generate(self, template_from, template_to,
                 parallel=True, backend="pystache"):
        """Same as `Experiments.generate`"""
        _do_generate_user(self, template_from, template_to,
                          parallel=parallel, backend=backend)

    def generate_jobs(self, system, template_to, export=[], depends=[],
                      submit_args=[], job_desc="jobs.jd", parallel=True):
        """Same as `Experiments.generate_jobs`"""
        _do_generate_jobs(self, system, template_to, export, depends,
                          submit_args, job_desc, parallel)

    def translate(self, template, with_exps=False, with_unique=True):
        """Same as `Experiments.translate`"""
        return _do_translate(self._base._experiments, self._filters, self._translate,
                             [template], with_exps, with_unique)


    def __getattr__(self, name):
        if name in ["_base", "_filters"]:
            object.__getattribute__(self, name)
        else:
            attr = getattr(self._base, name)
            if callable(attr):
                if attr.__name__ in VIEW_SUPPORTED:
                    attr = getattr(self._base,
                                   VIEW_SUPPORTED[attr.__name__])

                def func(*args, **kwargs):
                    return attr(self._as_filter(), *args, **kwargs)
                return func
            else:
                return attr

    def __setattr__(self, name, value):
        if name in ["_base", "_filters"]:
            object.__setattr__(self, name, value)
        else:
            attr = getattr(self._base, name)
            assert not callable(attr)
            setattr(self._base, name, value)


######################################################################
# Function wrappers

_THREADLOCAL = threading.local()

class _WithExp(pp.Pretty):

    def __init__(self, func):
        self._func = func

    @staticmethod
    def set_experiment(experiment):
        setattr(_THREADLOCAL, "_WithExp__experiment", experiment)

    @staticmethod
    def get_experiment():
        return getattr(_THREADLOCAL, "_WithExp__experiment", None)

    def __call__(self, *args, **kwargs):
        exp = self.get_experiment()
        return self._func(exp, *args, **kwargs)

    def _repr_pretty_(self, p, cycle):
        p.text("with_exp(")
        p.pretty(self._func)
        p.text(")")

    def __eq__(self, other):
        return isinstance(other, _WithExp) and self._func == other._func

    def __hash__(self):
        return hash((_WithExp, self._func))


def with_exp(func):
    """Wrap function to get experiment as first argument during translation.

    Examples
    --------
    >>> e = Experiments()
    >>> e.params(a=range(2), b=range(2),
    ...          c=with_exp(lambda exp: exp['a']+exp['b']))
    >>> e.translate("{{c}}")
    ['0', '1', '2']

    """
    return _WithExp(func)

class _WithExpTpl(pp.Pretty):

    _EXPERIMENT = None

    def __init__(self, func, template):
        self._func = func
        self._template = template
        self._translator = text.Translator(self._template)

    @staticmethod
    def set_experiment(experiment):
        setattr(_THREADLOCAL, "_WithExpTpl__experiment", experiment)

    @staticmethod
    def get_experiment():
        return getattr(_THREADLOCAL, "_WithExpTpl__experiment", None)

    def __call__(self, *args, **kwargs):
        exp = self.get_experiment()
        translated = self._translator.translate(exp)
        return self._func(exp, translated, *args, **kwargs)

    def _repr_pretty_(self, p, cycle):
        p.text("with_exp_tpl(")
        p.pretty(self._func)
        p.text(", ")
        p.pretty(self._template)
        p.text(")")

    def __eq__(self, other):
        return (isinstance(other, _WithExpTpl) and
                self._func == other._func and
                self._template == other._template)

    def __hash__(self):
        return hash((_WithExp, self._func, self._template))


def with_exp_tpl(func, template):
    """Similar to `with_exp`, but passes translated template as second argument.

    This is a convenience shortcut for using `text.translate` in `with_exp`.

    Examples
    --------
    In its basic form, we can combine translated templates with other experiment
    values:

    >>> e = Experiments()
    >>> def fun(exp, tpl):
    ...     return tpl + " == " + str(exp['a']+exp['b'])
    >>> e.params(a=range(2), b=range(2),
    ...          f=with_exp_tpl(fun, "{{a}}+{{b}}"))
    >>> e.translate("{{f}}")
    ['0+0 == 0', '0+1 == 1', '1+0 == 1', '1+1 == 2']

    But this comes in handy when we want to pull the contents of per-experiment
    files into new experiment values. Let's start by creating some files:

    >>> import tempfile
    >>> tmp_dir = tempfile.TemporaryDirectory()
    >>> with open(os.path.join(tmp_dir.name, "file-0"), "w") as f:
    ...     f.write("000")
    3
    >>> with open(os.path.join(tmp_dir.name, "file-1"), "w") as f:
    ...     f.write("111")
    3

    And now we can use them to populate experiment values:

    >>> def get_file(exp, path):
    ...    return open(path, "r").read()
    >>> v = e.view("b == 0")
    >>> v.params(f=with_exp_tpl(get_file, os.path.join(tmp_dir.name, "file-1")))
    >>> v.translate("{{f}}")
    ['111']
    >>> v.params(f=with_exp_tpl(get_file, os.path.join(tmp_dir.name, "file-{{a}}")))
    >>> v.translate("{{f}}")
    ['000', '111']

    """
    return _WithExpTpl(func, template)


class _FromFunction(object):
    def __init__(self, func, keys):
        self._func = func
        self._keys = keys
        self._results = None

    def param(self, key_name):
        if key_name not in self._keys:
            raise ValueError("unknown key name: %s" % key_name)
        if self._results is not None:
            raise Exception("cannot get parameter after getting function results")
        return _FromFunctionKey(self, key_name)

    def _get_results(self, exp):
        results = self._func(exp)
        if not isinstance(results, collections.abc.Sized):
            results = list(results)
        return results


class _FromFunctionKey(object):
    def __init__(self, from_function, key_name):
        self._from_func = from_function
        self._key_name = key_name


def from_function(fun, keys):
    """Parameter value to add given parameter permutations to every experiment.

    For every experiment, it will permute it with the values returned by
    `fun`. Note that `fun` is expected to return entries with multiple variable
    names (specified by `keys`), so you must call ``param`` on the result of
    `from_function` to select the specific variable name.

    Parameters
    ----------
    fun : callable
        Function to call on every experiment, receives the experiment at hand
        and must return a list of dict.
    keys : list of str
        Variable names in the elements returned by `fun`.

    Examples
    --------
    The result can use any of the values available on each experiment; in this
    case ``b`` is two and three times the value of ``a`` (note how `from_func`
    and other values can be mixed):

    >>> def func(exp):
    ...     return [{'b': exp["a"]*2}, {'b': exp["a"]*3}]
    >>> e = Experiments()
    >>> e.params(a=range(2), b=from_function(func, ["b"]).param("b"))
    >>> e
    Experiments([{'a': 0, 'b': 0}, {'a': 1, 'b': 2}, {'a': 1, 'b': 3}])

    See also
    --------
    Experiments.params

    """
    return _FromFunction(fun, keys)


def from_find_files(template, **kwargs):
    """Helper to use `sciexp2.common.utils.find_files` with `from_function`.

    Will provide all variable names that are tags in the template. If argument
    `path` is provided, will also provide that variable name (with the full path
    to the found file).

    Parameters
    ----------
    template : str
        Template of file paths to find.
    kwargs
        Arguments to `sciexp2.common.utils.find_files`.

    Examples
    --------
    After creating some example files:

    >>> import tempfile
    >>> tmp = tempfile.TemporaryDirectory()
    >>> _ = open(os.path.join(tmp.name, "0.txt"), "w")
    >>> _ = open(os.path.join(tmp.name, "1.txt"), "w")

    Now we can use the variables on the file template (``a``) and its full path
    (``PATH``) as new variables on the experiments (``var`` and ``path``):

    >>> e = Experiments()
    >>> f = from_find_files(os.path.join(tmp.name, "{{a}}.txt"), path="PATH")
    >>> e.params(var=f.param("a"), path=f.param("PATH"))
    >>> e                               # doctest:+ELLIPSIS
    Experiments([{'path': '/tmp/.../0.txt', 'var': 0},
                 {'path': '/tmp/.../1.txt', 'var': 1}])

    """
    template_vars = text.get_variables(template)
    keys = template_vars

    path = kwargs.get("path", None)
    if path is not None:
        if not isinstance(path, str):
            raise TypeError("path argument must be a string")
        keys.append(path)

    results = list(utils.find_files(template, **kwargs))

    def fun(experiment):
        experiment = dict((key, val) for key, val in experiment.items()
                          if key in template_vars)
        if len(experiment) > 0:
            filter_ = Filter(experiment)
        else:
            filter_ = Filter()
        for result in results:
            if filter_.match(result):
                yield result

    return from_function(fun, keys)


__all__ = [
    "Experiments", "ExperimentsView",
    "with_exp", "with_exp_tpl",
    "from_function", "from_find_files",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Common utility classes and functions."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2023, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import os
import shutil
import signal
import subprocess
import tempfile
import functools
import collections
import weakref
import numpy as np

from . import pp
from . import progress


# -----------------------------

def assert_kwargs(kwargs):
    """Raise an exception if extra keys are present."""
    if kwargs:
        extra = "s" if len(kwargs) > 1 else ""
        raise TypeError("Unexpected argument%s: %s" % (extra,
                                                       ", ".join(kwargs)))

# -----------------------------

def assert_dir(path):
    """Check that given directory exists, otherwise create it."""
    if path != "" and not os.path.exists(path):
        try:
            os.makedirs(path)
        except FileExistsError:
            # might be due to concurrent mkdir
            pass


def assert_path(path):
    """Check that given path exists, otherwise create directories."""
    if not path.endswith(os.sep):
        path = os.path.dirname(path)
    assert_dir(path)


def get_path(path):
    """Get path after expanding user and environment variables."""
    path = os.path.expanduser(path)
    return os.path.expandvars(path)


def get_file(path, mod="w"):
    """Open the given file, creating any intermediate directory."""
    dir_path = os.path.dirname(path)
    assert_dir(dir_path)
    return open(path, mod)


def get_tmp_file(mode="w", delete=True):
    """Get a temporal file."""
    return tempfile.NamedTemporaryFile(mode=mode, delete=delete)


with open(os.devnull, "w") as _null:
    _HAVE_RSYNC = subprocess.call(["which", "rsync"], stdout=_null)

if _HAVE_RSYNC == 0:
    def copy_path_rsync(path_from, path_to, preserve=True, dereference=False):
        """Copy contents using rsync."""
        if os.path.isdir(path_from):
            path_from = path_from + os.sep
            assert_path(path_to)
        else:
            assert_path(os.path.dirname(path_to) + os.sep)

        args = "-rptgoD"
        if preserve:
            args += "t"
        if not dereference:
            args += "l"
        else:
            args += "L"
        if subprocess.call(["rsync", args, path_from, path_to]) != 0:
            raise OSError("Error copying files: %s -> %s" % (
                path_from, path_to))

    def _copy_path(*args, **kwargs):
        copy_path_rsync(*args, **kwargs)

else:
    def copy_path_shutil(path_from, path_to, preserve=True, dereference=False):
        """Copy contents using Python's shutil."""
        if os.path.isdir(path_from):
            # NOTE: will fail if destination already exists
            path_from = path_from + os.sep
            assert_path(path_to)
            shutil.copytree(path_from, path_to, symlinks=not dereference)

        else:
            assert_path(os.path.dirname(path_to) + os.sep)
            if os.path.islink(path_from):
                link_to = os.readlink(path_from)
                os.symlink(link_to, path_to)
            else:
                shutil.copy(path_from, path_to)
            if preserve:
                shutil.copymode(path_from, path_to)

    def _copy_path(*args, **kwargs):
        copy_path_shutil(*args, **kwargs)


def copy_path(path_from, path_to, preserve=True, dereference=False):
    """Copy files."""
    _copy_path(path_from, path_to, preserve=preserve, dereference=dereference)

# -----------------------------

def str2num(arg):
    """Return numeric value of a string, if possible."""
    # NOTE: StringConverter barks at non-numeric strings (str/bytes confusion)
    try:
        return np.lib.npyio.StringConverter().upgrade(arg)
    except:
        return arg

# -----------------------------

def _wraps(wrapped):
    return functools.wraps(wrapped=wrapped,
                           assigned=['__doc__'])


class ViewError(Exception):
    """Invalid operation in `OrderedSet` view."""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class OrderedSet(collections.abc.MutableSet, pp.Pretty):
    """A mutable set preserving order of insertion.

    .. todo::

       All help should come from `~collections.abc.MutableSet` instead of
       using `_wraps`.

    """

    @_wraps(collections.abc.Container.__init__)
    def __init__(self, iterable=None, view_able=False):
        self._view_able = view_able
        if self._view_able:
            self._list = np.array([], dtype=object)
        else:
            self._list = []
        self._set_methods(False)
        self._base = None
        self._views = {}
        self._set = set()
        if iterable is not None:
            self |= iterable

    def set_view_able(self, view_able):
        """Set whether this object can produce "views" from it.

        Objects able to produce views have lower performance when adding new
        elements to them.

        See also
        --------
        OrderedSet.view

        """
        if view_able != self._view_able:
            if view_able:
                self._list = np.array(self._list, dtype=object)
            else:
                if self._views:
                    raise ValueError(
                        "cannot disable 'view_able' when views already exist")
                self._list = list(self._list)
            self._view_able = view_able
            self._set_methods(False)

    def view(self, index):
        """Create a view (sub-set) of this object.

        This object also becomes a view. Modifications to the elements of a view
        will also take effect on all other views of the same object.


        Parameters
        ----------
        index : slice


        See also
        --------
        OrderedSet.set_view_able

        """
        if not self._view_able:
            raise ValueError("the object is not 'view_able'")
        if not isinstance(index, slice):
            raise TypeError("view index must be a slice")

        self._set_methods(True)

        res = OrderedSet([], True)
        # pylint: disable=protected-access
        res._list = self._list[index]
        for elem in res._list:
            res._set.add(elem)
        res._base = self
        res._set_methods(True)

        self._views[id(res)] = weakref.ref(res)

        return res

    def __del__(self):
        if self._base is not None:
            # pylint: disable=protected-access
            del self._base._views[id(self)]

    def _set_methods(self, is_view):
        if self._view_able:
            if is_view:
                self._append = self._append_array_view
                self._remove = self._remove_array_view
                self._pop = self._pop_array_view
            else:
                self._append = self._append_array
                self._remove = self._remove_array
                self._pop = self._pop_array
        else:
            assert not is_view
            self._append = self._append_list
            self._remove = self._remove_list
            self._pop = self._pop_list

    def _append_list(self, value):
        self._list.append(value)

    def _remove_list(self, value):
        if self._base is not None:
            self._base.remove(value)
        else:
            self._list.remove(value)
            for view in self._views.values():
                # pylint: disable=protected-access
                view()._list.remove(value)
                view()._set.remove(value)

    def _pop_list(self, index):
        self._list.pop(index)

    def _append_array(self, value):
        self._list = np.append(self._list, value)

    def _remove_array(self, value):
        self._list = np.delete(self._list, np.where(self._list == value))

    def _pop_array(self, index):
        self._list = np.delete(self._list, index)

    # pylint: disable=no-self-use
    def _append_array_view(self, value):
        raise ViewError("cannot append to a view")

    # pylint: disable=no-self-use
    def _remove_array_view(self, value):
        raise ViewError("cannot remove from a view")

    # pylint: disable=no-self-use
    def _pop_array_view(self, index):
        raise ViewError("cannot pop from a view")

    # pylint: disable=invalid-name
    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(list(self._list))

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def get_index(self, index):
        """Get item at the 'index'th position."""
        return self._list[index]

    def copy(self):
        """Make a shallow copy of this `OrderedSet`."""
        return OrderedSet(self, self._view_able)

    def sorted(self, *args, **kwargs):
        """Same as `sort`, but returns a sorted copy."""
        res = self.copy()
        res.sort(*args, **kwargs)
        return res

    def sort(self, key=None, reverse=False):
        """Sort set in-place.

        Follows the same semantics of Python's built-in `sorted`.

        """
        if self._view_able:
            contents = list(self._list)
        else:
            contents = self._list

        contents.sort(key=key, reverse=reverse)

        if self._view_able:
            self._list[:] = contents

    # Container

    @_wraps(set.__contains__)
    def __contains__(self, key):
        return key in self._set

    # Sized

    @_wraps(set.__len__)
    def __len__(self):
        return len(self._list)

    # Iterable

    @_wraps(set.__iter__)
    def __iter__(self):
        return iter(self._list)

    # MutableSet

    # pylint: disable=missing-docstring
    def add(self, key):
        old_length = len(self._list)
        self._set.add(key)
        if len(self._set) != old_length:
            try:
                self._append(key)
            except ViewError:
                self._set.remove(key)
                raise
    add.__doc__ = collections.abc.MutableSet.add.__doc__

    # pylint: disable=missing-docstring
    def discard(self, key):
        old_length = len(self._list)
        self._set.remove(key)
        if len(self._set) != old_length:
            try:
                self._remove(key)
            except ViewError:
                self._set.add(key)
                raise
    discard.__doc__ = collections.abc.MutableSet.discard.__doc__
    discard.__doc__ += "\n\nThis operation has a cost of O(n)."

    # pylint: disable=missing-docstring
    @_wraps(collections.abc.MutableSet.pop)
    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self._pop(-1) if last else self._pop(0)
        self._set.remove(key)
        return key

    # Pickling

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict["_append"]
        del odict["_remove"]
        del odict["_pop"]
        del odict["_base"]
        del odict["_views"]
        return odict

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self._base = None
        self._views = {}
        self._set_methods(False)


# -----------------------------

def _template_get_initial_dir(template, template_is_abs):
    # pylint: disable=cyclic-import
    from . import text

    start_dir = ""
    for part in template.split(os.sep):
        if part == "":
            continue
        if start_dir == "" and not template_is_abs:
            cur_dir = part
        else:
            cur_dir = os.sep.join([start_dir, part])
        try:
            text.translate(cur_dir, {})
        except text.VariableError:
            break
        if os.path.isdir(cur_dir):
            start_dir = cur_dir
        else:
            break
    return start_dir


def find_files(template, path=None, absolute_path=False, sort=True):
    """Find files matching a given template.

    Returns an 'InstanceGroup' with all paths of existing files matching the
    given template. Each matching file path is an `Instance` with the extracted
    variables in the `template`.

    Parameters
    ----------
    template : str
        Template of file paths to find.
    path : str, optional
        On each resulting Instance, add a variable with the given name with the
        file path.
    absolute_path : bool, optional
        Make the value in `path` absolute.
    sort : bool, optional
        Sort the file paths according to the alphanumeric order of each of the
        variables in `template`, in that specific order.

    Raises
    ------
    ValueError
        The variable in `path` is already present in `template`.

    See Also
    --------
    sciexp2.common.text.extract
        Argument `template` is interpreted following the extraction syntax.

    Notes
    -----
    If `template` ends with ``/`` it will search for matching paths, and will
    search for matching files otherwise.

    Environment variables and user home directories in `template` will be expanded.

    """
    # pylint: disable=cyclic-import
    from . import text
    from .instance import InstanceGroup, Instance

    if not isinstance(template, str):
        raise ValueError("Not an expression: " + template)
    if path is not None and not isinstance(path, str):
        raise ValueError("path must be either None or a string")
    if path in text.get_variables(template):
        raise ValueError("path variable is already present in template")

    template_is_dir = template[-1] == "/" if template else False
    template_is_abs = os.path.isabs(template)
    template = get_path(template) + "$"
    start_dir = _template_get_initial_dir(template, template_is_abs)

    extractor = text.Extractor(template)
    res = InstanceGroup()

    def add(env, target_path):
        # use numbers whenever possible (for later number-aware sorting)
        for key, val in env.items():
            env[key] = str2num(val)
        if path is not None:
            if absolute_path:
                target_path = os.path.abspath(target_path)
            env[path] = target_path
        res.add(Instance(env))

    for dir_path, _, file_list in os.walk(start_dir):
        if template_is_dir:
            try:
                env = extractor.extract(dir_path + os.path.sep)
            except text.ExtractError:
                pass
            else:
                add(env, dir_path)
        else:
            for file_path in file_list:
                target_path = os.path.join(dir_path, file_path)
                try:
                    env = extractor.extract(target_path)
                except text.ExtractError:
                    pass
                else:
                    add(env, target_path)

    if sort:
        # sort result according to file sorting
        variables = text.get_variables(template)
        res.sort(variables)

    return res


# -----------------------------

def execute_with_sigint(cmd, **kwargs):
    """Execute a command and forward SIGINT to it.

    Parameters
    ----------
    cmd : list of string
        Command to execute
    kwargs : dict
        Additional arguments to subprocess.Popen.

    Returns
    -------
    Integer with the command's return code.

    """
    preexec_fn = kwargs.pop("preexec_fn", None)
    def preexec():
        os.setpgrp()
        if preexec_fn:
            preexec_fn()

    signals = [
        ("SIGINT", signal.SIGINT),
        ("SIGTERM", signal.SIGTERM),
        ("SIGKILL", signal.SIGKILL),
    ]

    state = dict(proc=None,
                 error=False,
                 signal_idx=0)

    def run():
        if state["proc"] is None:
            if not state["error"]:
                # pylint: disable=subprocess-popen-preexec-fn
                state["proc"] = subprocess.Popen(cmd, preexec_fn=preexec, **kwargs)
            else:
                return
        state["proc"].wait()

    def run_with_except(depth=0):
        try:
            run()
        except KeyboardInterrupt:
            state["error"] = True
            info = signals[state["signal_idx"]]
            progress.log(progress.LVL_NONE,
                         "WARNING: Interrupting child command with %s" % info[0])
            state["proc"].send_signal(info[1])
            if state["signal_idx"] < len(signals) - 1:
                state["signal_idx"] += 1
            run_with_except(depth + 1)
            if depth == 0:
                raise

    run_with_except()
    return state["proc"].returncode



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Text translation and extraction functions.

The translation and extraction functions are based on text templates that follow
a subset of the mustache syntax: https://mustache.github.io.

The accepted subset of mustache tags is:

* Variables: ``{{name}}``
* Sections: ``{{#name}}...{{/name}}``
* Inverted sections: ``{{^name}}...{{/name}}``

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2019-2023, 2025, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import ast
from collections import OrderedDict
try:
    from collections.abc import Mapping
except:
    pass
import functools
import jinja2
import os
import pystache
import re
from .utils import OrderedSet
import sys


class ParseError(Exception):
    pass


class VariableError(Exception):
    pass


class ExtractError(Exception):
    pass


def _get_parsed_elems(parsed):
    return parsed._parse_tree


@functools.lru_cache(maxsize=4096)
def _pystache_parse(text, allow_nested=True, allow_inverted=True):
    try:
        pystache.render(text, {})
    except pystache.parser.ParsingError as e:
        _, _, exc_traceback = sys.exc_info()
        new_e = ParseError(str(e))
        raise new_e.with_traceback(exc_traceback)

    parsed = pystache.parse(text)

    elems = _get_parsed_elems(parsed)
    if len(elems) == 0 and len(text) > 0:
        raise ParseError("section start tag mismatch")

    def traverse(elems, nested):
        seen_node = False
        for elem in elems:
            if not isinstance(elem, str):
                seen_node = True

            if isinstance(elem, str):
                pass
            elif isinstance(elem, (pystache.parser._EscapeNode,
                                   pystache.parser._ChangeNode)):
                pass
            elif isinstance(elem, pystache.parser._SectionNode):
                nested = traverse(_get_parsed_elems(elem.parsed), True)
                if nested and not allow_nested:
                    raise ParseError(
                        "nested tags not allowed in section %r" % elem.key)
            elif isinstance(elem, pystache.parser._InvertedNode):
                if not allow_inverted:
                    raise ParseError("inverted sections not allowed: %s" % elem.key)
                nested = traverse(_get_parsed_elems(elem.parsed_section), True)
                if nested and not allow_nested:
                    raise ParseError(
                        "nested tags not allowed in inverted section %r" % elem.key)
            elif isinstance(elem, pystache.parser._PartialNode):
                raise ParseError(
                    "partial tags not allowed")
            else:
                raise ParseError("tag not allowed %r" % elem.__class__)

        return seen_node

    traverse(elems, False)

    return parsed


def get_variables(text, nested=False):
    """Get the variables referenced in the given text.

    Parameters
    ----------
    text : str
       Text to get variables from.
    nested : optional
       Whether to allow nested variables. Can have values "all" for all the
       variables, or "inner" for just the inner variables.

    Examples
    --------
    >>> get_variables("{{a}}")
    ['a']
    >>> get_variables("{{#a}} {{b}} {{/a}}", nested="inner")
    ['b']
    >>> get_variables("{{#a}} {{b}} {{/a}}", nested="all")
    ['a', 'b']

    """
    if nested not in [False, "all", "inner"]:
        raise ValueError("invalid nested value:", nested)

    parsed = _pystache_parse(text, allow_nested=bool(nested))
    if not nested:                    # equivalent due to exception raised above
        nested = "all"

    def traverse(elems, variables):
        added_variables = False

        for elem in elems:
            if isinstance(elem, pystache.parser._SectionNode):
                traversed_variables = traverse(_get_parsed_elems(elem.parsed),
                                               variables)
                if nested == "all":
                    variables.add(elem.key)
                    added_variables = True
                elif nested == "inner" and not traversed_variables:
                    variables.add(elem.key)
                    added_variables = True

            elif isinstance(elem, pystache.parser._InvertedNode):
                traversed_variables = traverse(_get_parsed_elems(elem.parsed_section),
                                               variables)
                if nested == "all":
                    variables.add(elem.key)
                    added_variables = True
                elif nested == "inner" and not traversed_variables:
                    variables.add(elem.key)
                    added_variables = True

            elif isinstance(elem, (pystache.parser._EscapeNode, pystache.parser._PartialNode)):
                variables.add(elem.key)
                added_variables = True

            else:
                assert isinstance(elem, str), elem

        return added_variables

    elems = _get_parsed_elems(parsed)
    variables = set()
    traverse(elems, variables)

    return sorted(variables)


class TemplatePath:
    """Wrapper to identify a path to a template file."""
    def __init__(self, path):
        self.value = path


class TemplateContents:
    """Wrapper"""
    def __init__(self, contents):
        self.value = contents


class Translator(object):
    """Translate a template text with given variables."""

    DEFAULT_BACKEND = "pystache"

    def __init__(self, template, backend=None, backend_options={}):
        """
        Parameters
        ----------
        template : str or TemplateContents or TemplatePath
            Template contents (when using `TemplateContents`) or path to a file
            containing them (when using `TemplatePath`). Passing a `str` is a
            shorthand for `TemplateContents(template)`.
        backend : {'pystache', 'jinja2'}, optional
            Name of the templating backend to process the contents of `template`.
        backend_options : dict, optional
            Options for the translation backend.

        Notes
        -----
        The "pystache" backend recognizes no options.

        The "jinja2" backend recognizes the following options:
        - `loader_search_path` : list of str, default [os.getcwd(), "/"]
            Paths to search for template and any files it includes. Uses the
            jinja2.FileSystemLoader() to find files.

        """
        if isinstance(template, str):
            template = TemplateContents(template)
        assert isinstance(template, tuple([TemplatePath, TemplateContents]))
        if backend is None:
            backend = Translator.DEFAULT_BACKEND
        backend_options = dict(backend_options)

        if backend == "pystache":
            if len(backend_options):
                raise RuntimeError("unknown backend option: " + list(backend_options.keys())[0])

            def identity(arg):
                return arg
            if isinstance(template, TemplatePath):
                template_txt = open(template.value).read()
            elif isinstance(template, TemplateContents):
                template_txt = template.value
            else:
                assert False
            parsed_template = _pystache_parse(template_txt, allow_nested=True)
            renderer = pystache.renderer.Renderer(search_dirs=[], file_extension=False,
                                                  partials=None, escape=identity,
                                                  missing_tags="strict")
            def render(env, recursive):
                template_new = template
                parsed_new = parsed_template
                while True:
                    template_old = template_new
                    parsed_old = parsed_new
                    try:
                        template_new = renderer.render(parsed_new, env)
                    except pystache.context.KeyNotFoundError as e:
                        _, _, exc_traceback = sys.exc_info()
                        new_e = VariableError("missing variable %s" % e.key)
                        raise new_e.with_traceback(exc_traceback)
                    except pystache.common.TemplateNotFoundError as e:
                        _, _, exc_traceback = sys.exc_info()
                        new_e = VariableError(str(e))
                        raise new_e.with_traceback(exc_traceback)

                    if not recursive:
                        break
                    elif template_old == template_new:
                        break

                    parsed_new = _pystache_parse(template_new, allow_nested=True)

                return template_new

            self._render = render

        elif backend == "jinja2":
            loader_search_path = backend_options.pop("loader_search_path",
                                                     [os.getcwd(), "/"])
            environment_filters = backend_options.pop("environment_filters", {})
            if len(backend_options):
                raise RuntimeError("unknown backend option: " + list(backend_options.keys())[0])

            def get_template(template):
                if isinstance(template, TemplatePath):
                    jinja_env = jinja2.Environment(
                        loader=jinja2.FileSystemLoader(loader_search_path),
                        undefined=jinja2.StrictUndefined,
                    )
                    jinja_env.filters.update(environment_filters)
                    jinja_tpl = jinja_env.get_template(template.value)
                elif isinstance(template, TemplateContents):
                    jinja_env = jinja2.Environment(
                        loader=jinja2.DictLoader({"template": template.value}),
                        undefined=jinja2.StrictUndefined,
                    )
                    jinja_env.filters.update(environment_filters)
                    jinja_tpl = jinja_env.get_template("template")
                else:
                    assert False
                return jinja_tpl

            def render(env, recursive):
                template_new = template
                parsed_new = template
                jinja_tpl = get_template(template_new)

                while True:
                    template_old = template_new
                    parsed_old = parsed_new

                    try:
                        template_new = jinja_tpl.render(env)
                    except jinja2.exceptions.UndefinedError as e:
                        _, _, exc_traceback = sys.exc_info()
                        new_e = VariableError(e.message)
                        raise new_e.with_traceback(exc_traceback)

                    if not recursive:
                        break
                    elif template_old == template_new:
                        break

                    jinja_tpl = get_template(TemplateContents(template_new))

                return template_new

            self._render = render

        else:
            raise RuntimeError("invalid templating backend: " + str(backend))

    def translate(self, env, recursive=True):
        """Apply translation with given variables.

        Parameters
        ----------
        env : dict
            Mapping of variable names to their values.
        recursive : bool, optional
            Whether to apply translations recursively.

        Examples
        --------
        You can perform simple text translations:

        >>> t = Translator('Hello {{a}}')
        >>> t.translate({'a': 'you'})
        'Hello you'
        >>> t.translate({'a': [1, 2]})
        'Hello [1, 2]'

        And also recursive ones:

        >>> t.translate({'a': '{{b}}', 'b': 'them'})
        'Hello them'

        More complex cases like conditionals are also possible:

        >>> t = Translator('{{#a}}is true{{/a}}{{^a}}is false{{/a}}')
        >>> t.translate({'a': 1})
        'is true'
        >>> t.translate({'a': 0})
        'is false'

        Or even calls to functions (arguments are the unexpanded text on the template):

        >>> Translator('{{a}}').translate({'a': lambda: 'value'})
        'value'
        >>> Translator('{{#a}}{{b}}{{/a}}').translate(
        ...     {'a': lambda arg: 2*arg, 'b': 4})
        '44'
        >>> Translator('{{#a}}{{b}}{{/a}}').translate(
        ...     {'a': lambda arg: " ".join(list(arg))})
        '{ { b } }'


        And expansion of nested variables with multiple values is also possible:

        >>> Translator('{{#a}}A.B=={{b}} {{/a}}').translate({'a': [{'b': 1}, {'b': 2}]})
        'A.B==1 A.B==2 '

        """
        if not isinstance(env, Mapping):
            raise TypeError("not a mapping: %r" % env)

        return self._render(env, recursive)

    def translate_many(self, envs, recursive=True, ignore_variable_error=False,
                       with_envs=False):
        """Apply translation with given set of variables.

        Parameters
        ----------
        envs : sequence of dict
            Sequence of variable names to value mappings to apply the
            translation for.
        recursive : bool, optional
            Whether to apply translations recursively.
        ignore_variable_error : bool, optional
            Ignore translations for variable maps that have missing variables.
        with_envs : bool, optional
            Get the set of maps that led to each translation.

        Returns
        -------
        list of str
            Translations when ``with_envs`` is ``False``.
        list of (str, [env])
            Translations with their corresponding variable maps when
            ``with_envs`` is ``True``.

        Notes
        -----
        The result is guaranteed to maintain the order of the elements of
        `envs`.

        Examples
        --------
        You can very easily translate a sequence of variable maps:

        >>> t = Translator('Hello {{a}}')
        >>> t.translate_many([{'a': 'you'}, {'a': 'them'}])
        ['Hello you', 'Hello them']

        Multiple maps can also translate into the same text:

        >>> t.translate_many([{'a': 'you'}, {'a': 'them', 'b': 1}, {'a': 'them', 'b': 2}])
        ['Hello you', 'Hello them']

        But you can also get the maps that led to each translation:

        >>> t = Translator('Hello {{a}}')
        >>> translated = t.translate_many([{'a': 'you'}, {'a': 'them', 'b': 1},
        ...                                {'a': 'them', 'b': 2}], with_envs=True)
        >>> translated == [('Hello you', [{'a': 'you'}]),
        ...                ('Hello them', [{'a': 'them', 'b': 1},
        ...                                {'a': 'them', 'b': 2}])]
        True

        """
        if with_envs:
            result = OrderedDict()
            def add(key, val):
                if key not in result:
                    result[key] = []
                result[key].append(val)
        else:
            result_track = OrderedSet()
            result = []
            def add(key, val):
                if key not in result_track:
                    result_track.add(key)
                    result.append(key)

        for env in envs:
            try:
                text = self.translate(env)
            except VariableError:
                if not ignore_variable_error:
                    raise
            else:
                add(text, env)

        if with_envs:
            return list(result.items())
        else:
            return result


def translate(template, env, **kwargs):
    """Shorthand for ``Translator(template).translate(env, **kwargs)``."""
    backend = kwargs.pop("backend", None)
    backend_options = kwargs.pop("backend_options", {})
    translator = Translator(template=template, backend=backend,
                            backend_options=backend_options)
    return translator.translate(env=env, **kwargs)


def translate_many(template, envs, **kwargs):
    """Shorthand for ``Translator(template).translate_many(envs, **kwargs)``."""
    backend = kwargs.pop("backend", None)
    backend_options = kwargs.pop("backend_options", {})
    translator = Translator(template=template, backend=backend,
                            backend_options=backend_options)
    return translator.translate_many(envs=envs, **kwargs)


class Extractor(object):
    """Extract a dict with the variable values that match a given template.

    Variables and sections on the template are used to define regular
    expressions, following Python's `syntax
    <http://docs.python.org/library/re.html#regular-expression-syntax>`_.

    """

    def __init__(self, template):
        """
        Parameters
        ----------
        template : str
            Template text to extract from.

        """
        self._template = template
        parsed = _pystache_parse(template, allow_nested=False, allow_inverted=False)
        regex = ""
        variables = {}
        for elem in _get_parsed_elems(parsed):
            if isinstance(elem, str):
                regex += elem

            elif isinstance(elem, pystache.parser._SectionNode):
                if elem.key in variables:
                    raise ParseError(
                        "regex for variable %s has already been set: %s" % (
                            elem.key, variables[elem.key]))
                elem_regex = _get_parsed_elems(elem.parsed)
                if len(elem_regex) == 0:
                    raise ParseError(
                        "regex for variable %s cannot be empty" % elem.key)
                elem_regex = elem_regex[0]
                assert len(elem_regex) > 0, template
                variables[elem.key] = elem_regex
                regex += "(?P<%s>%s)" % (elem.key, elem_regex)

            elif isinstance(elem, pystache.parser._EscapeNode):
                if elem.key in variables:
                    regex += "(?P=%s)" % elem.key
                else:
                    elem_regex = ".+"
                    variables[elem.key] = elem_regex
                    regex += "(?P<%s>%s)" % (elem.key, elem_regex)

            else:
                # silently ignore
                pass

        self._cre = re.compile(regex)

    def extract(self, text):
        """Apply extraction to given text.

        Parameters
        ----------
        text : str
            Text to extract from.

        Examples
        --------
        You can perform simple text extractions, where variables correspond to
        the simple regex ``.+``:

        >>> e = Extractor('Hello {{a}}')
        >>> e.extract('Hello world')
        {'a': 'world'}
        >>> e.extract('Hello 123!')
        {'a': '123!'}

        More complex regexes can be specified using section tags:

        >>> Extractor('Hello {{#a}}[0-9]+{{/a}}.*').extract('Hello 123!')
        {'a': 123}

        And using the same variable on multiple tags ensures they all match the
        same contents:

        >>> extracted = Extractor('{{#a}}[0-9]+{{/a}}.*{{a}}{{b}}').extract('123-123456')
        >>> extracted == {'a': 123, 'b': 456}
        True

        """
        match = self._cre.match(text)
        if match is None:
            raise ExtractError(
                "could not extract variables from template %r (regex: %r)" % (
                    self._template, self._cre.pattern))
        def retype_maybe(val):
            try:
                return ast.literal_eval(val)
            except:
                pass
            return val
        return dict((key, retype_maybe(val))
                    for (key, val) in match.groupdict().items())


def extract(template, text):
    """Shorthand for ``Extractor(template).extract(text)``."""
    return Extractor(template).extract(text)


__all__ = [
    "ParseError", "VariableError", "ExtractError",
    "get_variables",
    "TemplatePath", "TemplateContents",
    "Translator", "translate", "translate_many",
    "Extractor", "extract",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Progress indicators.

Progress messages are only shown if the provided level is greater than the one
in `level`.

The available levels are: `LVL_NONE`, `LVL_PROGRESS` (the default), `LVL_INFO`,
`LVL_VERBOSE` and `LVL_DEBUG`.

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2024, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections
import itertools
import numbers
import os
import sys
import threading
import types

import tqdm


LVL_NONE = 0                      #: No progress indication.
LVL_PROGRESS = 1                  #: Show progress indication.
LVL_INFO = 2                      #: Show additional information.
LVL_VERBOSE = 3                   #: Be verbose when progressing.
LVL_DEBUG = 4                     #: Show debugging information.

_LEVEL = LVL_PROGRESS

if "ipykernel" in sys.modules:
    # we're in a notebook ...
    try:
        import ipywidgets.widgets.widget
        # pylint: disable=protected-access
        if getattr(ipywidgets.widgets.widget.Widget, "_version_validated", None) is not None:
            # ... and widgets are usable
            _CLS = tqdm.tqdm_notebook
        else:
            _CLS = tqdm.tqdm
    #pylint: disable=bare-except
    except:
        _CLS = tqdm.tqdm
else:
    _CLS = tqdm.tqdm


def get_writer(file):
    read_fd, write_fd = os.pipe()

    write_wrapper = os.fdopen(write_fd, "w")
    read_wrapper = os.fdopen(read_fd, "r")

    def reader():
        while True:
            data = read_wrapper.readline()
            if len(data) == 0:
                return
            tqdm.tqdm.write(data, end="", file=file)
    reader_thread = threading.Thread(target=reader, daemon=True)
    reader_thread.start()

    return write_wrapper


def level(level_=None):
    """Get/set the current progress indication level."""
    if level_ is not None:
        if level_ < LVL_NONE or LVL_DEBUG < level_:
            raise ValueError(f"Invalid progress level {level_}")
        # pylint: disable=global-statement
        global _LEVEL
        _LEVEL = level_
    return _LEVEL


def _log(fmt, *args):
    with _PRINT_LOCK:
        tqdm.tqdm.write(fmt % (args), file=sys.stderr)

_PRINT_LOCK = threading.Lock()


def log(level_, fmt, *args, **kwargs):
    """Log message with given level."""
    if level_ > _LEVEL:
        return
    _log(fmt, *args, **kwargs)


def info(fmt, *args, **kwargs):
    """Log message with 'info' level."""
    log(LVL_INFO, fmt, *args, **kwargs)


def verbose(fmt, *args, **kwargs):
    """Log message with 'verbose' level."""
    log(LVL_VERBOSE, fmt, *args, **kwargs)


def debug(fmt, *args, **kwargs):
    """Log message with 'debug' level."""
    log(LVL_DEBUG, fmt, *args, **kwargs)


class Progress:
    """A progress indicator.

    Parameters
    ----------
    total : optional
        Objective progress count, or sized object.
    msg : string, optional
        Message to prefix to the progress indicator.

    """

    def __init__(self, total=None, msg=""):
        total = _get_total(total)
        if level() < LVL_PROGRESS:
            self._obj = None
        else:
            self._obj = _CLS(desc=msg, total=total, leave=False, dynamic_ncols=True)

    def __call__(self, increment=1):
        """Increment the progress indicator by the given number."""
        if self._obj is not None:
            self._obj.update(increment)

    def __iadd__(self, other):
        """Increment the total count of the progress indicator."""
        if self._obj.total is None:
            self._obj.total = 0
        self._obj.total += other
        return self


    def _cleanup(self):
        if self._obj is not None:
            self._obj.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._cleanup()

    def __del__(self):
        """Delete the progress indicator from screen."""
        self._cleanup()

    def write(self, s, file=None, end="\n", nolock=False):
        if self._obj is not None:
            return self._obj.write(s, file=file, end=end, nolock=nolock)
        else:
            if file is None:
                file = sys.stdout
            return file.write(s, end=end)


class Null:
    """A "null" progress indicator.

    """

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def __iadd__(self, other):
        return self

    def _cleanup(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def __del__(self):
        pass

    def write(self, s, file=None, end="\n", nolock=False):
        if file is None:
            file = sys.stdout
        return file.write(s, end=end)


def _progressable_get_func_wrapper(func, other, progr):
    """Update progress every time `func` is called."""
    # pylint: disable=unused-argument
    def wrapper(self, *args, **kwargs):
        res = func(other, *args, **kwargs)
        progr()
        return res
    return wrapper


def _progressable_get_iter_wrapper(func, other, progr):
    """Update progress every time the result of `func` is iterated."""
    # pylint: disable=unused-argument
    def wrapper(self, *args, **kwargs):
        otheriter = func(other, *args, **kwargs)

        counter = set()

        class _ProgressableIterator(other.__class__):
            def __init__(self):
                pass

            def __iter__(self):
                return self

            # pylint: disable=missing-docstring
            def next(self):
                return next(self)

            # pylint: disable=no-self-use
            def __next__(self):
                res = next(otheriter)
                if counter:
                    # must offset progress by one element
                    # (returned element does not finish until we retrieve next)
                    progr()
                else:
                    counter.add(None)
                return res

            def __getattr__(self, attr):
                try:
                    return getattr(otheriter, attr)
                # pylint: disable=bare-except
                except:
                    return getattr(other, attr)

            def __setattr__(self, attr, val):
                setattr(otheriter, attr, val)

        return _ProgressableIterator()
    return wrapper


# pylint: disable=too-few-public-methods
class Progressable:
    """Indicates objects returned by `progressable`."""


def progressable(other, progr, funcs=None, iters=None):
    """Get a wrapper that invokes a progress indicator on some routines.

    Parameters
    ----------
    other
        Object to wrap.
    progr
        A progress indicator instance.
    funcs : list of str, optional
        List of routines in `other` to wrap.
    iters : list of str, optional
        List of routines in `other` that return an interator to wrap.

    Notes
    -----
    Argument `other` is also passed as the first argument to `get`, such that
    it can return a proper indicator.

    """
    if funcs is None:
        funcs = []
    if iters is None:
        iters = []
    assert funcs or iters
    assert len(funcs) + len(iters) == len(set(funcs) | set(iters))

    class _Progressable(other.__class__, Progressable):
        def __init__(self):
            pass

        def __enter__(self):
            progr.__enter__()
            return self

        def __exit__(self, *args, **kwargs):
            progr.__exit__(*args, **kwargs)

        def __getattr__(self, attr):
            return getattr(other, attr)

        def __setattr__(self, attr, val):
            return other.__class__.__setattr__(other, attr, val)

    res = _Progressable()

    for func in itertools.chain(funcs, iters):
        if not hasattr(other, func):
            raise ValueError(f"Routine {func} not found")
        func_orig = getattr(other.__class__, func)
        if func in funcs:
            func_ = _progressable_get_func_wrapper(func_orig, other, progr)
        else:
            func_ = _progressable_get_iter_wrapper(func_orig, other, progr)
        setattr(_Progressable, func, func_)

    return res


def progressable_simple(contents, length, *args, **kwargs):
    """A simplified version of `progressable`.

    Creates a progress indicator from the given arguments.

    """
    if length is None:
        progr = get(contents, *args, **kwargs)
    else:
        progr = get(length, *args, **kwargs)

    if isinstance(contents, types.GeneratorType):
        # pylint: disable=missing-docstring
        class Wrapper(collections.abc.Iterable, collections.abc.Sized):
            def __iter__(self):
                return iter(contents)

            def __len__(self):
                return len(contents)

        res = Wrapper()
    else:
        res = contents

    return progressable(res, progr, iters=["__iter__"])


class Stack:
    """Stack of currently active progress indicators.

    Provides only static attributes and methods, in order to provide a
    program-wide stateful stack of progress indicators.
    """

    _STACK = []
    CURRENT = None
    """The currently active progress indicator."""

    @staticmethod
    def __call__(*args, **kwargs):
        """Call on the current (latest) progress indicator.

        Parameters
        ----------
        args, kwargs
            Arguments for ``__call__`` to the current progress indicator.
        """
        # pylint: disable=not-callable
        Stack.CURRENT(*args, **kwargs)

    @staticmethod
    def push(klass, *args, **kwargs):
        """Add a new progress indicator to the stack and make it current.

        Parameters
        ----------
        klass : callable
            Class object or function to construct the new progress indicator.
        args, kwargs
            Arguments to `klass`.
        """
        if not callable(klass):
            raise TypeError

        if level() < LVL_PROGRESS:
            obj = Null(*args, **kwargs)
        else:
            obj = klass(*args, **kwargs)

        Stack._STACK.append(obj)
        Stack.CURRENT = Stack._STACK[-1]
        return Stack.CURRENT

    @staticmethod
    def pop():
        """Remove the last progress indicator from the stack."""
        assert len(Stack._STACK) > 1
        res = Stack.CURRENT
        Stack._STACK.pop()
        Stack.CURRENT = Stack._STACK[-1]
        return res

# Add a default null progress reporter
Stack.push(Null)


class UnpicklerStart:
    """Picklable proxy to a progress indicator.

    In order to provide a progress indicator when unpickling large objects, you
    can prefix the pickle stream with a `UnpicklerStart` instance and postfix
    it with its `UnpicklerStop` sister instance::

        def dump (self, file_obj):
            start = UnpicklerStart(len(self))
            pickle.dump((start, self, start.get_stop()), file_obj)

    When unpickled, the first will call `Stack.push` and the last will call
    `Stack.pop`, so that your object will be able to update the
    `~Stack.CURRENT` progress indicator by calling `Stack`.

    When pickled, it will also push a progress indicator for your code to use
    while pickling your large object.

    Parameters
    ----------
    obj
        Object to show the progress indicator for
    no_pickle : bool, optional
        Do not create a progress indicator while pickling
    msg_pickle : string, optional
        Message during pickling (otherwise use default)
    no_unpickle : bool, optional
        Do not create a progress indicator while unpickling
    msg_unpickle : string, optional
        Message during unpickling (otherwise use default)togg

    """

    def __init__(self, obj, **kwargs):
        no_pickle = kwargs.pop("no_pickle", False)
        msg_pickle = kwargs.pop("msg_pickle", "saving to file")
        no_unpickle = kwargs.pop("no_unpickle", False)
        msg_unpickle = kwargs.pop("msg_unpickle", "loading from file")
        # pylint: disable=cyclic-import
        # pylint: disable=import-outside-toplevel
        from . import utils
        utils.assert_kwargs(kwargs)

        total = _get_total(obj)

        if no_pickle:
            self._pickle = None
        else:
            cls = _get_class()
            self._pickle = (cls, total, msg_pickle)

        if no_unpickle:
            self._unpickle = None
        else:
            self._unpickle = (Progress, total, msg_unpickle)

    def get_stop(self):
        """Get the sister `UnpicklerStop` object."""
        return UnpicklerStop(self._pickle is not None,
                             self._unpickle is not None)

    def __getstate__(self):
        if self._pickle is not None:
            Stack.push(*self._pickle)
        contents = self.__dict__.copy()
        del contents["_pickle"]
        return contents

    def __setstate__(self, contents):
        self.__dict__.update(contents)
        if self._unpickle is not None:
            cls, total, msg = self._unpickle
            cls = _get_class()
            Stack.push(cls, total, msg)


class UnpicklerStop:
    """Remove the progress indicators created by `UnpicklerStart`."""

    def __init__(self, do_pickle, do_unpickle):
        """
        Parameters
        ----------
        do_pickle, do_unpickle : boolean
            Remove a progress indicator after pickling / unpickling.
        """
        self._pickle = do_pickle
        self._unpickle = do_unpickle

    def __getstate__(self):
        contents = self.__dict__.copy()
        del contents["_pickle"]
        if self._pickle:
            progress = Stack.pop()
            #pylint: disable=protected-access
            progress._cleanup()
        return contents

    def __setstate__(self, contents):
        self.__dict__.update(contents)
        if self._unpickle:
            progress = Stack.pop()
            #pylint: disable=protected-access
            progress._cleanup()


def _get_class():
    if level() >= LVL_PROGRESS:
        return Progress
    return Null


def _get_total(obj):
    if isinstance(obj, str):
        return None
    if isinstance(obj, collections.abc.Sized):
        return len(obj)
    if isinstance(obj, numbers.Number):
        return obj
    return None


def get(obj, msg=""):
    """Return a progress indicator.

    """
    cls = _get_class()
    return cls(obj, msg)


def get_pickle(obj, **kwargs):
    """Return an object to show a progress indicator during (un)pickling.

    The progress indicator can be reached through Stack.CURRENT.

    Parameters
    ----------
    obj
        Object to show the progress indicator for
    kwargs
        Additional arguments for UnpicklerStart.

    """
    return UnpicklerStart(obj, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Execution system and job abstractions.

Each execution system consists of a Python module with:

* a subclass of `System` named ``System``.

* a subclass of `Job` that ``System`` will use.

Execution systems are Python modules, and searched for using Python's module
search paths (see PYTHONPATH).

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2024, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import abc
import glob
import importlib
import os
import shutil
import weakref

import sciexp2.common.instance
from sciexp2.common.filter import *
from sciexp2.common import text
from sciexp2.common import utils


_DEVNULL = open("/dev/null", "w")


class SystemError (Exception):
    """Error loading system."""
    def __init__(self, message):
        Exception.__init__(self, message)


class SubmitArgsError (Exception):
    """Error translating job submission arguments."""
    def __init__(self, variables):
        Exception.__init__(
            self,
            "Found non-exported variables in job submission arguments: " +
            ", ".join(variables))


class System (abc.ABC):
    """Abstract job manager.

    Each system must implement the abstract methods defined in this class
    and define two class attributes:

        ========= =======================================================
        Name      Description
        ========= =======================================================
        `ASSUMES` List of variables that are assumed to be present in the
                  launchers instance group for the system to work.
        `DEFINES` List of variables that the system internally defines and thus
                  must not be present in the launchers instance group.
        ========= =======================================================

    See also
    --------
    compute_state

    """

    ASSUMES = ["LAUNCHER", "DONE", "FAIL"]
    DEFINES = ["_STATE", "LAUNCHER_BASE"]

    def __init__(self, base_path, launchers, depends, submit_args):
        """
        Parameters
        ----------
        base_path : str
            Base directory where launchers are located.
        launchers : InstanceGroup
            Group describing the launchers.
        depends : sequence of str
            Variable names to which jobs depend.
        submit_args : sequence of str
            Extra arguments to the job-submitting program.

        """
        self._base = base_path
        self._base_abs = os.path.realpath(self._base)
        assert os.path.isdir(self._base_abs)
        self._launchers = launchers
        for assume in self.assumes():
            if assume not in self._launchers and len(self._launchers) > 0:
                raise ValueError("Variable '%s' must be present" % assume)
        for define in self.defines():
            if define in self._launchers:
                raise ValueError("Variable '%s' must not be present" % define)
        self._jobs = None
        self._depends = set(depends)
        self._submit_args = list(submit_args)

    def get_relative_path(self, path, cwd=None):
        """Get path (relative to base) as relative to `cwd`."""
        if cwd is None:
            cwd = os.getcwd()
        if not os.path.isabs(path):
            path = os.path.join(self._base_abs, path)
        return os.path.relpath(path, cwd)

    def build(self, types, *filters):
        """Generate a sequence with the jobs matching the given criteria.

        Parameters
        ----------
        types : set
            Set of states that the jobs must be on.
        filters : list of filters
            List of filters that the jobs must match.

        See also
        --------
        Job

        """
        self.compute_state()

        build_filter = and_filters(*filters)

        if len(types) > 0:
            state_filter = " or ".join(["_STATE == '%s'" % state
                                      for state in types
                                      if state != "inverse"])
            if "inverse" in types:
                state_filter = "not (%s)" % state_filter
            build_filter = and_filters(build_filter, state_filter)

        if len(self._jobs) > 0:
            build_filter.validate(set(self._jobs.variables()))
            return self._jobs.select(build_filter)
        else:
            return sciexp2.common.instance.InstanceGroup()

    @classmethod
    def assumes(cls):
        """The set of variables that must be present on the launchers."""
        return set(System.ASSUMES + cls.ASSUMES)

    @classmethod
    def defines(cls):
        """The set of variables that must not be present on the launchers."""
        return set(System.DEFINES + cls.DEFINES)

    @abc.abstractmethod
    def compute_state(self):
        """Compute the current state of jobs.

        The implementation must set the ``_jobs`` attribute with an
        InstanceGroup of `Job` instances. This can be computed using the
        contents of the ``_launchers`` attribute.

        """
        pass

    @staticmethod
    def post_generate(base, path, instance, xlator):
        """Post-process the generation of file `path`."""
        pass


class Job (sciexp2.common.instance.Instance, abc.ABC):
    """Abstract job descriptor.

    Each job must implement the abstract methods defined in this class.

    See also
    --------
    state, submit, kill

    """

    # job states
    RUNNING = "running"
    DONE = "done"
    FAILED = "failed"
    OUTDATED = "outdated"
    NOTRUN = "notrun"

    STATES = [
        RUNNING,
        DONE,
        FAILED,
        OUTDATED,
        NOTRUN,
    ]

    STATE_SHORT = {
        RUNNING:  u"\u2699",
        DONE:     u"\u2713",
        FAILED:   "x",
        OUTDATED: "o",
        NOTRUN:   " ",
        }

    STATE_LONG = {
        RUNNING:  "Running",
        DONE:     "Done",
        FAILED:   "Failed",
        OUTDATED: "Outdated",
        NOTRUN:   "Not run",
        }

    def __init__(self, system, state, instance):
        """
        Parameters
        ----------
        system : System
            System for which this job is.
        state : str
            Execution state of the job.
        instance : str
            Launcher instance describing this job.

        """
        sciexp2.common.instance.Instance.__init__(self, instance)
        self["_STATE"] = state
        self._system = weakref.proxy(system)

    def __repr__(self):
        return repr(sciexp2.common.instance.Instance(self))

    @classmethod
    def compute_state(cls, system, instance):
        """Generic job state computation.

        Parameters
        ----------
        system : System
            System for which this job is being checked.
        instance
            Launcher instance describing a job.

        Returns
        -------
        Generic job state according to the failed/done files; otherwise returns
        `NOTRUN`.

        """
        fail_path = instance["FAIL"]
        if not os.path.isabs(fail_path):
            fail_path = os.sep.join([system._base, fail_path])
        if os.path.exists(fail_path):
            return cls.FAILED

        done_path = instance["DONE"]
        if not os.path.isabs(done_path):
            done_path = os.sep.join([system._base, done_path])
        if not os.path.exists(done_path):
            return cls.NOTRUN

        done_mtime = os.stat(done_path).st_mtime
        for dep in system._depends:
            path = text.translate(dep, instance)
            if path == "":
                continue
            path = utils.get_path(path)
            if not os.path.isabs(path):
                path = os.sep.join([system._base, path])
            if not os.path.exists(path) or \
               done_mtime < os.stat(path).st_mtime:
                return cls.OUTDATED
        return cls.DONE

    @abc.abstractmethod
    def state(self):
        """Return a string describing the job and its state."""
        pass

    @abc.abstractmethod
    def submit(self, *args):
        """Submit a job to execution."""
        pass

    def _submit_args(self, args):
        """Return extra arguments for the job submitting program."""
        instance = dict(self)
        instance["LAUNCHER_BASE"] = self._system._base_abs
        try:
            return [text.translate(arg, instance)
                    for arg in self._system._submit_args + list(args)]
        except text.VariableError as e:
            raise SubmitArgsError(e.message)

    @abc.abstractmethod
    def kill(self, *args):
        """Kill a job from execution."""
        pass

    def _kill_args(self, args):
        """Return extra arguments for the job killing program."""
        instance = dict(self)
        instance["LAUNCHER_BASE"] = self._system._base_abs
        try:
            return [text.translate(arg, instance)
                    for arg in list(args)]
        except text.VariableError as e:
            raise SubmitArgsError(e.message)


def get(name):
    """Get an execution system implementation by name."""
    try:
        system = importlib.import_module(name)
    except ImportError:
        try:
            system = importlib.import_module("sciexp2.expdef.system." + name)
        except ImportError:
            raise SystemError("Unknown system %r" % name)
    try:
        res = system.System
    except AttributeError:
        raise AttributeError("Does not look like an execution " +
                             "system implementation: %s" % name)
    res.name = name
    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Support for instances and associated abstractions."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2024, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import collections
import contextlib
import pickle
import weakref

from .filter import Filter
from . import progress
from . import utils
from . import pp


_EXPORT = [set()]

_INVALID = object()


class Instance(dict, pp.Pretty):
    """A map from variable names to values."""

    def __init__(self, source=None):
        """
        Parameters
        ----------
        source : :class:`dict` or ``None``
            Initial values for this instance.

        """
        if source is None:
            dict.__init__(self)
        else:
            dict.__init__(self, source)
        setattr(self, "__translate_cache", {})
        setattr(self, "__translate_cache_enabled", False)

    @contextlib.contextmanager
    def auto_translate(self):
        setattr(self, "__translate_cache_enabled", True)
        yield
        setattr(self, "__translate_cache_enabled", False)

    def __getitem__(self, key):
        if not getattr(self, "__translate_cache_enabled"):
            return dict.__getitem__(self, key)

        cache = getattr(self, "__translate_cache")
        val = cache.get(key, _INVALID)
        if val is _INVALID:
            val = dict.__getitem__(self, key)
            if isinstance(val, str):
                from . import text
                try:
                    val = text.translate(val, self, recursive=True)
                except text.VariableError:
                    pass
            cache[key] = val
        return val

    def __setitem__(self, key, val):
        getattr(self, "__translate_cache").clear()
        return dict.__setitem__(self, key, val)

    def __hash__(self):
        """Hashing function for adding into `InstanceGroup`."""
        return id(self)

    def __iter__(self):
        """Get an iterator to ``(key, value)`` pairs in the Instance."""
        return iter(dict.items(self))

    def copy(self):
        """Return a shallow copy of self."""
        return Instance(self)

    def deepcopy(self, *keys):
        """Return a copy of self, deepcopying the values of given keys."""
        res = Instance(self)
        for key in keys:
            res[key] = self[key].__class__(self[key])
        return res

    # pylint: disable=invalid-name
    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            with p.group(1, "{", "}"):
                for idx, (var, val) in enumerate(sorted(self.items())):
                    if idx:
                        p.text(",")
                        p.breakable()
                    p.pretty(var)
                    p.text(": ")
                    # p.breakable()
                    p.pretty(val)

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def __getstate__(self):
        # pylint: disable=not-callable
        progress.Stack.CURRENT()
        if not _EXPORT[-1]:
            odict = self.__dict__
        else:
            odict = {}
            for key, val in self.items():
                if key in _EXPORT[-1]:
                    odict[key] = val
        return odict

    def __reduce__(self):
        # pylint: disable=not-callable
        progress.Stack.CURRENT()
        if not _EXPORT[-1]:
            odict = dict(self)
        else:
            odict = {}
            for key in _EXPORT[-1]:
                val = self.get(key, _INVALID)
                if val is not _INVALID:
                    odict[key] = val
        return tuple([Instance, tuple(), odict])

    def __setstate__(self, contents):
        # pylint: disable=not-callable
        progress.Stack.CURRENT()
        self.update(contents)


class ViewError(Exception):
    """Invalid operation in an `InstanceGroup` view."""
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class InstanceGroup(pp.Pretty):
    """A group of `Instance` objects suitable for searching.

    Provides a fast way to retrieve a set of Instance objects that have certain
    variables set or certain variable values.

    Parameters
    ----------
    instances : iterable of Instance objects, optional
        Initial contents. Contents are copied during initialization.
    cache, cache_proactive, cache_others
        Arguments to `cache_set`.
    view_able : bool, optional
        Whether to enable the use of `view`.

    See also
    --------
    cache_set

    Notes
    -----
    .. warning::

       You should **not** modify the results of accessing an InstanceGroup,
       unless you know what you're doing.

       As this operation returns the same data being used internally, changes
       to it must be kept consistent with the cache (see `cache_reset`).

    .. warning::

       You should **not** add or extend a group with instances that have the
       same contents as other instances already present in the group.

       .. todo:: An `InstanceGroup` should mimic an `OrderedSet`

    Examples
    --------
    >>> g = InstanceGroup([{'a':4, 'b':3}, {'a':2, 'b':1}, {'a':4}])
    >>> g
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4})])

    Testing the existence of variables:

    >>> 'a' in g
    True
    >>> 'c' in g
    False

    Getting all the possible values for a variable:

    >>> g['a']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict({4: OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})]),
                 2: OrderedSet([Instance({'a': 2, 'b': 1})])})
    >>> g['b']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict({3: OrderedSet([Instance({'a': 4, 'b': 3})]),
                 1: OrderedSet([Instance({'a': 2, 'b': 1})])})

    Getting all the instances with a specific value:

    >>> g['a'][4]                          # doctest: +NORMALIZE_WHITESPACE
    OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})])

    Extending a group with another:

    >>> h = InstanceGroup([{'a':1}, {'a':2}])
    >>> h
    InstanceGroup([Instance({'a': 1}), Instance({'a': 2})])
    >>> g + h
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4}),
                   Instance({'a': 1}),
                   Instance({'a': 2})])
    >>> g += h
    >>> g
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4}),
                   Instance({'a': 1}),
                   Instance({'a': 2})])
    >>> g['a']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict({4: OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})]),
                 2: OrderedSet([Instance({'a': 2, 'b': 1}), Instance({'a': 2})]),
                 1: OrderedSet([Instance({'a': 1})])})

    """

    def __init__(self, instances=None, cache=True,
                 cache_proactive=False, cache_others=None,
                 view_able=False):
        self._view_able = view_able
        self._base = None
        self._views = {}
        self._instances = utils.OrderedSet(view_able=self._view_able)
        self._reverse = dict()

        self._cache = True
        self._cache_proactive = False
        self._cache_others = None
        self.cache_set(cache, cache_proactive, cache_others)

        if instances is not None:
            for i in instances:
                self.add(Instance(i))

    def set_view_able(self, view_able):
        """Set whether this object can produce "views" from it.

        Objects able to produce views have lower performance when adding new
        elements to them.


        See also
        --------
        InstanceGroup.view

        """
        self._instances.set_view_able(view_able)
        self._view_able = view_able

    def view(self, index):
        """Create a view (sub-set) of this object.

        This object also becomes a view. Modifications to the elements of a view
        will also take effect on all other views of the same object.


        Parameters
        ----------
        index : slice


        See also
        --------
        InstanceGroup.set_view_able

        """
        if not self._view_able:
            raise ValueError("the object is not 'view_able'")
        if not isinstance(index, slice):
            raise TypeError("view index must be a slice")

        res = InstanceGroup([], cache=self._cache,
                            cache_proactive=self._cache_proactive,
                            cache_others=self._cache_others, view_able=True)
        # pylint: disable=protected-access
        res._instances = self._instances.view(index)
        res._reverse = dict(self._reverse)
        res.cache_reset()
        res.add = self._add_view
        self.add = self._add_view
        res._base = self
        self._views[id(self)] = weakref.ref(res)
        return res

    def copy(self):
        """Return a copy."""
        return InstanceGroup(self,
                             cache=self._cache,
                             cache_proactive=self._cache_proactive,
                             cache_others=self._cache_others,
                             view_able=self._view_able)

    def get_index(self, index):
        """Get instance at the  `index`'th position."""
        return self._instances.get_index(index)

    # pylint: disable=no-self-use,unused-argument,method-hidden
    def add(self, instance):
        """Add an `Instance` into this group.

        The caching state will have an impact on the performance of this
        operation.

        See also
        --------
        cache_set

        """
        # actual function set by `_set_cache_proactive`
        assert False

    def _set_cache_proactive(self, proactive, others=None):
        if isinstance(proactive, bool):
            if self._cache_proactive:
                self.add = self._add_cache_all
            else:
                self.add = self._add_cache_none

        else:
            if not proactive and others is None:
                self.add = self._add_cache_direct
            elif others is None:
                proactive = set(proactive)
                self.add = self._add_cache_list
            else:
                for var in others:
                    self._cache_reset(var)
                proactive = tuple(proactive)
                if not proactive:
                    self.add = self._add_cache_direct
                else:
                    self.add = self._add_cache_list_only

            prefetch = [var for var in proactive
                        if var in set(self.variables())]
            self.cache_prefetch(prefetch, skip_cached=False)

        self._cache_proactive = proactive
        self._cache_others = others

    # pylint: disable=no-self-use
    def _add_view(self, instance):
        raise ViewError("cannot add to a view")

    def _add_cache_direct(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)

    def _add_cache_all(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for key, val in instance:
            self._cache_add(key, val, instance)

    def _add_cache_none(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        # just initialize existence of variables
        for key, _ in instance:
            self._cache_reset(key)

    def _add_cache_list(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for key, val in instance:
            if key in self._cache_proactive:
                self._cache_add(key, val, instance)
            else:
                self._cache_reset(key)

    def _add_cache_list_only(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for k in self._cache_proactive:
            self._cache_add(k, instance[k], instance)

    def __iadd__(self, obj):
        """Extend this group with one more instance or group.

        .. todo::

           Should probably delete extension with instances.

        """
        try:
            self.add(obj)
        # pylint: disable=broad-except
        except Exception:
            if not isinstance(obj, Instance):
                self.extend(obj)
            else:
                raise
        return self

    def extend(self, instances):
        """Extend this group with the contents of another group."""
        assert isinstance(instances, InstanceGroup)
        for instance in instances:
            self.add(instance)

    def __add__(self, instances):
        """x.__add__(y) <==> x + y

        Equivalent to a new `InstanceGroup` extended with both groups.

        """
        res = InstanceGroup(self)
        res.extend(instances)
        return res

    def __contains__(self, variable):
        """Syntax sugar for `has_variable`."""
        return variable in self._reverse

    def has_variable(self, variable):
        """Whether the group contains any Instance with given variable name."""
        return variable in self._reverse

    def has_instance(self, instance):
        """Whether this group contains the given Instance."""
        return instance in self._instances

    def variables(self):
        """The sequence of variables in this group."""
        return self._reverse.keys()

    def __getitem__(self, var):
        """Get all values for the given variable."""
        res = self._reverse[var]
        if res is None:  # JIT caching
            res = collections.OrderedDict()
            for instance in self._instances:
                try:
                    val = instance[var]
                except KeyError:
                    continue
                try:
                    vals = res[val]
                except KeyError:
                    vals = utils.OrderedSet()
                    res[val] = vals
                vals.add(instance)
            if self._cache:
                self._reverse[var] = res
        return res

    def _cache_add(self, key, value, instance):
        try:
            keys = self._reverse[key]
        except KeyError:
            keys = collections.OrderedDict()
            self._reverse[key] = keys

        try:
            vals = keys[value]
        except TypeError:               # 'key' uncached
            keys = collections.OrderedDict()
            self._reverse[key] = keys
            vals = utils.OrderedSet()
            keys[value] = vals
        except KeyError:                # 'value' not present
            vals = utils.OrderedSet()
            keys[value] = vals

        vals.add(instance)

    def _cache_reset(self, key):
        self._reverse[key] = None

    def cache_get(self):
        """Get the caching state."""
        return self._cache

    def cache_set(self, enable, proactive=None, others=None):
        """Set the caching status for this group.

        Caching offers greater speed of getting a specific variable multiple
        times (e.g., ``g['a']``), at the expense of a higher memory footprint.

        Parameters
        ----------
        enable : bool
           Caching status. When the caching is disabled, automatically calls
           `cache_reset`.
        proactive : bool or sequence of variable names, optional
           Proactively cache slected variables when adding new instances
           (default is to keep last value).
        others : sequence of variable names, optional
           List of other variables that the group does/will contain (default is
           to keep last value).

        See also
        --------
        cache_get, cache_reset, cache_prefetch

        Notes
        -----
        If caching is disabled, every call to `add` will reset the cache for
        the variables of the added instance.

        If `proactive` is `True`, it will proactively cache all variables, if
        it is `False` it will not. If it is a sequence of variable names, only
        these variables will be proactively cached.

        You can also provide an empty `proactive` list and an appropriate
        `others` list to get zero-caching during instance addition, which is
        the fastest option available.

        .. warning ::

           Providing `others` will greatly enhance the performance of `add`,
           but the contents will be inconsistent if `others` does not contain
           exactly all the variables of the instances previously present or
           added before interacting with the group.

           You can use `cache_reset` with the `hard` argument to ensure
           consistency.

        """
        assert isinstance(enable, bool)
        if proactive is not None:
            if enable is False:
                raise ValueError("Cannot use proactive caching when "
                                 "caching is disabled")
            if not isinstance(proactive, bool) and (
                    not isinstance(proactive, collections.abc.Iterable) or
                    not all([isinstance(elem, str) for elem in proactive])):
                raise TypeError("Invalid `proactive` caching "
                                "value: %s" % proactive)
        if others is not None:
            if proactive is None:
                raise ValueError(
                    "Cannot use 'others' without proactive caching")
            if not isinstance(others, bool) and (
                    not isinstance(others, collections.abc.Iterable) or
                    not all([isinstance(elem, str) for elem in others])):
                raise TypeError("Invalid `others` caching "
                                "value: %s" % others)

        if self._cache != enable:
            self._cache = enable
            if not enable:
                self.cache_reset()

        if proactive is not None:
            self._set_cache_proactive(proactive, others)

    def cache_reset(self, hard=False):
        """Reset the cache of this group.

        Parameters
        ----------
        hard : bool optional
            Perform a costly "hard" cache reset by inspecting all instances in
            the group.

        See also
        --------
        cache_get, cache_set

        """
        if hard:
            self._reverse = dict()
            for instance in self._instances:
                for var in instance.keys():
                    self._cache_reset(var)
        else:
            for var in self._reverse.keys():
                self._cache_reset(var)

    def cache_prefetch(self, variables=None, skip_cached=True):
        """Enable caching and force it for the given variables.

        Parameters
        ----------
        variables : sequence of variable names, optional
            Variables to prefetch. Uses all variables when ``None``.
        skip_cached : bool, optional
            Do not recompute caching if a variable is already cached.

        See also
        --------
        cache_get, cache_set

        Notes
        -----
        When prefetching a variable that is already cached, not necessarily all
        its values will be present on the cache (only those previously
        referenced); thus the `skip_cached` argument.

        """
        self._cache = True

        # select variables
        if variables is None:
            variables = self.variables()
        else:
            self_variables = list(self.variables())
            for var in variables:
                if var not in self_variables:
                    raise ValueError("No such variable: '%s'" % var)

        # uncache variables to be cached
        variables = set(variables)
        for var in set(variables):
            if self._reverse[var] is not None and skip_cached:
                variables.remove(var)
            else:
                self._cache_reset(var)

        # start caching
        if not variables:
            return
        for instance in self._instances:
            for var in variables:
                if var not in instance:
                    continue
                self._cache_add(var, instance[var], instance)

    def __len__(self):
        """The number of instances in this group."""
        return len(self._instances)

    def __iter__(self):
        """An iterator to the instances in the group."""
        return iter(self._instances)

    def select(self, filter_, allow_unknown=False):
        """Generate a sequence of instances matching the given filter.

        Parameters
        ----------
        filter_ : filter
            Filter establishing which instances to select from this group
            (accepts anything convertible into a filter).
        allow_unknown : bool or callable, optional
            Whether to allow the filter to reference variables not present in
            all instances of the group (in which case the instance is not
            selected). If a callable is used, it will receive the instance and
            its result will be used.

        """
        filter_ = Filter(filter_)
        def generate(instances):
            for instance in instances:
                try:
                    match = filter_.match(instance)
                except NameError:
                    if callable(allow_unknown):
                        if not allow_unknown(instance):
                            raise
                    elif not bool(allow_unknown):
                        raise
                else:
                    if match:
                        yield instance
        return generate(self._instances)

    def sorted(self, *args, **kwargs):
        """Same as `sort`, but returns a sorted copy."""
        res = self.copy()
        res.sort(*args, **kwargs)
        return res

    def sort(self, variables, key=None):
        """Sort group in-place by the given variables.

        Parameters
        ----------
        variables : list of strings
            Variable names for which to sort their values. The instances are
            sorted according to the order of the given variable names.
        key : func, optional
            Function returning a value that will be used for comparison. By
            default compares all instance variables as a list.

        See also
        --------
        functools.cmp_to_key

        """
        # get list of all variables in group
        variables = list(variables)
        for var in self.variables():
            if var not in variables:
                variables.append(var)

        if key is None:
            # transform values into a list, sorted according to `variables`
            def to_list(instance):
                return tuple(instance[var] for var in variables)
            key = to_list

        self._instances.sort(key=key)

    # pylint: disable=invalid-name
    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(list(self._instances))

    #######################################################################
    # pickling
    #######################################################################

    def dump(self, file_obj, export=None):
        """Write a pickled representation.

        Parameters
        ----------
        file_obj : file
            The destination file.
        export : set of strings, optional
            Variables that must be exported. Defaults to all.

        See also
        --------
        pickle

        """
        if export is None:
            export = set()
        _EXPORT.append(set(export))
        start = progress.get_pickle(self)
        stop = start.get_stop()
        pickle.dump((start, self, stop), file_obj, -1)
        _EXPORT.pop()

    def __getstate__(self):
        odict = self.__dict__.copy()
        new_reverse = dict()
        for key in odict["_reverse"]:
            if not _EXPORT[-1] or key in _EXPORT[-1]:
                new_reverse[key] = None
        odict["_reverse"] = new_reverse
        if "add" in odict:
            del odict["add"]
        del odict["_base"]
        del odict["_views"]
        return odict

    @staticmethod
    def load(file_obj):
        """Load a pickled representation.

        See also
        --------
        pickle

        """
        _, res, _ = pickle.load(file_obj)
        assert isinstance(res, InstanceGroup)
        return res

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self._base = None
        self._views = {}
        self._set_cache_proactive(self._cache_proactive, self._cache_others)


__all__ = [
    "Instance",
    "ViewError", "InstanceGroup",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Functions typically used with `~sciexp2.expdef`:

- `~sciexp2.expdef.experiments.Experiments`
- `~sciexp2.expdef.experiments.with_exp`
- `~sciexp2.expdef.experiments.with_exp_tpl`
- `~sciexp2.expdef.experiments.from_function`
- `~sciexp2.expdef.experiments.from_find_files`
- `~sciexp2.common.filter.Filter`
- `~sciexp2.common.filter.and_filters`
- `~sciexp2.common.filter.or_filters`
- `~sciexp2.common.text.TemplateContents`
- `~sciexp2.common.text.TemplatePath`

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013-2025, Lluís Vilanova"
__license__ = "GPL version 3 or later"


__all__ = []


def _env_add(name, value):
    __all__.append(name)
    globals()[name] = value


def _env_setup():
    from . import experiments
    import sciexp2.common.filter

    _env_add("Experiments", experiments.Experiments)
    _env_add("with_exp", experiments.with_exp)
    _env_add("with_exp_tpl", experiments.with_exp_tpl)
    _env_add("from_function", experiments.from_function)
    _env_add("from_find_files", experiments.from_find_files)
    _env_add("Filter", sciexp2.common.filter.Filter)
    _env_add("and_filters", sciexp2.common.filter.and_filters)
    _env_add("or_filters", sciexp2.common.filter.or_filters)
    _env_add("TemplateContents", sciexp2.common.text.TemplateContents)
    _env_add("TemplatePath", sciexp2.common.text.TemplatePath)

_env_setup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__     = "Lluís Vilanova"
__copyright__  = "Copyright 2008-2026, Lluís Vilanova"
__license__    = "GPL version 3 or later"


__version_info__ = (2, 2, 1)
__version__ = ".".join([str(i) for i in __version_info__])

__all__ = [
    "env", "experiments", "launcher", "system", "templates",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import sys
import setuptools

base = os.path.dirname(sys.argv[0])

version_file = os.path.join(base, "sciexp2", "expdef", "__init__.py")
with open(version_file, encoding="utf-8") as f:
    code = compile(f.read(), version_file, 'exec')
    version = {}
    exec(code, {}, version) #, global_vars, local_vars)

with open(os.path.join(base, "README.md")) as f:
    readme = f.read()
long_description = re.match(r"[\w-]+\n=+\n(.*)\nCopyright$\n=========$", readme,
                            re.MULTILINE | re.DOTALL).group(1)
long_description = long_description.strip()

with open(os.path.join(base, ".requirements.txt"), "r") as f:
    reqs = [line[:-1] for line in f.readlines()]

setuptools.setup(
    name="sciexp2-expdef",
    version=version["__version__"],
    description="Experiment definition framework for SciExp²",
    long_description=long_description,
    author="Lluís Vilanova",
    author_email="llvilanovag@gmail.com",
    url="https://sciexp2-expdef.readthedocs.io/",
    license="GNU General Public License (GPL) version 3 or later",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
    ],
    packages=["sciexp2", "sciexp2.common", "sciexp2.expdef", "sciexp2.expdef.system"],
    package_data={"sciexp2.expdef": ["templates/*.dsc", "templates/*.tpl"]},
    scripts=["launcher"],
    install_requires=reqs,
    platforms="any",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Common interface to manage jobs created with expdef.

Each job is described with an `~sciexp2.common.instance.Instance`, as generated
by `~sciexp2.expdef.experiments.Experiments.launcher`.

You can integrate job management into your application by instantiating the
`Launcher` class.

"""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2023, Lluís Vilanova"
__license__ = "GPL version 3 or later"


import hashlib
import os
import stat
import struct
import time
import pickle
import subprocess

from sciexp2.common.filter import *
from sciexp2.common import instance
from sciexp2.common import progress
from sciexp2.common import text
from sciexp2.common import utils
import sciexp2.expdef.system



class Launcher:
    """Common interface to manage jobs.

    The `types` argument to all methods identifies a set of job states to
    narrow the selection of jobs.

    The `filters` argument to all methods narrows the selection of jobs to
    those matching the given filters.

    """

    def __init__(self, base_path, system, group, depends, submit_args):
        """
        Parameters
        ----------
        base_path : str
            Path to the base directory created by
            `~sciexp2.expdef.experiments.Experiments.launcher`.
        system : str or System subclass
            Execution system.
        group : `~sciexp2.common.instance.InstanceGroup`
            Job descriptors.
        depends : sequence of str
            Variable names to which jobs depend.
        submit_args : sequence of str
            Extra arguments to the job-submitting program.

        See also
        --------
        sciexp2.expdef.system.get

        """
        if not os.path.isdir(base_path):
            raise IOError("Not a directory: " + base_path)

        if isinstance(system, str):
            name = system
            system = sciexp2.expdef.system.get(system)
            if system is None:
                raise ValueError("Not a launcher system name: " + name)
        if not issubclass(system, sciexp2.expdef.system.System):
            raise TypeError("Not a System subclass: " +
                            system.__class__.__name__)

        if not isinstance(group, instance.InstanceGroup):
            raise TypeError("Not an InstanceGroup")

        self._system = system(base_path, group, depends, submit_args)

    def parse_filters(self, *filters):
        value_filters = []
        filters_ = []
        for f in filters:
            try:
                f = Filter(f)
            except SyntaxError:
                path_abs = os.path.abspath(f)
                path = self._system.get_relative_path(
                    path_abs, self._system._base_abs)
                found = None
                for var in self.variables():
                    values = self._system._launchers[var]
                    if f in values:
                        found = "%s == %r" % (var, f)
                        break
                    if path in values:
                        found = "%s == %r" % (var, path)
                        break
                    if path_abs in values:
                        found = "%s == %r" % (var, path_abs)
                        break
                if found is None:
                    raise
                value_filters.append(found)
            else:
                filters_.append(f)
        res = and_filters(*filters_)
        if len(value_filters) > 0:
            res = and_filters(res, or_filters(*value_filters))
        res.validate(self.variables())
        return res

    def variables(self):
        """Return a list with the available variables."""
        return list(self._system._launchers.variables())

    def values(self, *filters):
        """Return a list with the available variables."""
        filter_ = and_filters(*filters)
        return list(self._system._launchers.select(filter_))

    def summary(self, types, *filters):
        """Print a summary of the state of selected jobs."""
        if len(types) == 0:
            types = list(sciexp2.expdef.system.Job.STATES)

        parts = [(sciexp2.expdef.system.Job.STATE_LONG[t],
                  len(list(self._system.build([t], *filters))))
                 for t in types]
        max_name = max(len(t) for t, _ in parts)
        max_count = max(len(str(n)) for _, n in parts)
        fmt = "%%-%ds: %%%dd" % (max_name, max_count)
        for t, n in parts:
            print(fmt % (t, n))

    def state(self, types, *filters, **kwargs):
        """Print the states for the selected jobs."""
        expression = kwargs.pop("expression", None)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))

        for job in self._system.build(types, *filters):
            state, name = job.state()
            if expression is not None:
                name = text.translate(expression, job)
            print("(%s) %s" % (sciexp2.expdef.system.Job.STATE_SHORT[state], name))

    def submit(self, types, *filters, **kwargs):
        """Submit selected jobs to execution.

        Calls `kill` before submitting a job if it's already running.

        """
        submit_args = kwargs.pop("submit_args", [])
        keep_going = kwargs.pop("keep_going", False)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))
        jobs = list(self._system.build(types, *filters))
        exit_code = 0
        with progress.progressable_simple(
                jobs, None, msg="Submitting jobs...") as pjobs:
            for job in pjobs:
                if job["_STATE"] == sciexp2.expdef.system.Job.RUNNING:
                    job.kill()
                job_path = self._system.get_relative_path(job["LAUNCHER"])
                progress.info("Submitting %s", job_path)
                try:
                    job.submit(*submit_args)
                except subprocess.CalledProcessError as e:
                    exit_code = e.returncode
                    msg = "ERROR: Exited with code %d" % exit_code
                    if progress.level() < progress.LVL_INFO:
                        msg += " (%s)" % self._system.get_relative_path(job["LAUNCHER"])
                    progress.log(progress.LVL_NONE, msg)
                    if not keep_going:
                        break
        return exit_code

    def kill(self, types, *filters, **kwargs):
        """Kill the execution of selected jobs."""
        kill_args = kwargs.pop("kill_args", [])
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))
        for job in self._system.build(types, *filters):
            if job["_STATE"] == sciexp2.expdef.system.Job.RUNNING:
                job.kill(*kill_args)

    def files(self, types, *filters, **kwargs):
        """List files matching an expression."""
        expression = kwargs.pop("expression")
        not_expanded = kwargs.pop("not_expanded", False)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))

        expr_path = expression
        if not os.path.isabs(expr_path):
            expr_path = os.path.join(self._system._base_abs, expr_path)

        expr_files = list(f["__FILE"]
                          for f in utils.find_files(expr_path,
                                                    path="__FILE"))

        job_files = set()
        for job in self._system.build(types, *filters):
            path = text.translate(expression, job)
            abs_path = path
            if not os.path.isabs(abs_path):
                abs_path = os.path.join(self._system._base_abs, abs_path)
            if not os.path.exists(abs_path):
                continue
            job_files.add(abs_path)

        if not_expanded:
            res = set(expr_files) - job_files
        else:
            res = set(expr_files) & job_files
        for i in sorted(res, key=expr_files.index):
            print(i)


def _header():
    return b"#!/usr/bin/env launcher\n"


def _magic():
    version = b"launcher"
    return hashlib.md5(version).digest()


class LauncherLoadError(Exception):
    """Could not load given file."""

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "Not a job descriptor file: %s" % self.path


def save(file_name, base_to_file, system, group, export, depends, submit_args):
    """Save an InstanceGroup as a job descriptor into a file.

    Parameters
    ----------
    file_name : str
        Path to destination file.
    base_to_file : str
        Relative path from some base directory to the directory containing
        `file_name`.
    system : `~sciexp2.expdef.system.System`
        Class of the execution system to use.
    group : `~sciexp2.common.instance.InstanceGroup`
        Job descriptors.
    export : set
        Variable names to export into the description file.
    depends : sequence of str
        Variable names to which jobs depend.
    submit_args : sequence of str
        Extra arguments to the job-submitting program.

    See also
    --------
    load, sciexp2.expdef.system.get

    """
    assert system is not None
    if not issubclass(system, sciexp2.expdef.system.System):
        system = sciexp2.expdef.system.get(system)

    if len(group) > 0:
        for assumed in system.assumes():
            if assumed not in group:
                raise ValueError("'%s' must be defined" % assumed)

    file_dir = os.path.dirname(file_name)
    if not os.path.exists(file_dir):
        os.makedirs(file_dir)
    with open(file_name, "wb") as file_obj:
        file_obj.write(_header())

        version = _magic()
        file_obj.write(struct.pack("I", len(version)))
        file_obj.write(version)

        pickle.dump((base_to_file, system.name, depends, submit_args),
                    file_obj, -1)
        group.dump(file_obj, export)
    fmode = stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH
    os.chmod(file_name, fmode)


def load(file_name):
    """Load a job description from a file.

    Parameters
    ----------
    file_name : str
        Path to source file.

    Returns
    -------
    sciexp2.expdef.system.System
        Instance of the execution system with the job descriptions.

    Raises
    ------
    LauncherLoadError
        The given file cannot be loaded as a `Launcher`.

    See also
    --------
    save
    """
    with open(file_name, "rb") as file_obj:
        header = file_obj.readline()
        if header != _header():
            raise LauncherLoadError(file_name)
        version_size = struct.unpack("I", file_obj.read(struct.calcsize("I")))
        version_size = version_size[0]
        version = file_obj.read(version_size)
        if version != _magic():
            raise LauncherLoadError(file_name)

        base_to_file, system_name, depends, submit_args = pickle.load(file_obj)
        group = instance.InstanceGroup.load(file_obj)

    jd_base = os.path.dirname(os.path.abspath(file_name))
    if base_to_file != "":
        base = jd_base[:-len(base_to_file)]
    else:
        base = jd_base
    base = os.path.relpath(base)
    return Launcher(base, system_name, group, depends, submit_args)
